# do imports here so I can do a single line import
from pathlib import Path
from textwrap import dedent
import math
import random
import time
from copy import copy, deepcopy
from math import floor, ceil, inf

from ursina.window import instance as window
from ursina.camera import instance as camera
from ursina.mouse import instance as mouse
from ursina.scene import instance as scene
from ursina.main import Ursina
from panda3d.core import Mat3, Mat4
from ursina.ursinamath import *
from ursina.destroy import destroy
from ursina.ursinastuff import *
from ursina.array_tools import *
from ursina import input_handler
from ursina.input_handler import held_keys, Keys
from ursina.string_utilities import *
from ursina.mesh_importer import load_model, load_blender_scene
from ursina.texture import Texture
from ursina.texture_importer import load_texture
from ursina import color
from ursina.color import Color, hsv, rgb
from ursina.sequence import Sequence, Func, Wait
from ursina import curve
from ursina.entity import Entity
from ursina.collider import *
from ursina.raycast import raycast
from ursina.boxcast import boxcast
from ursina.audio import Audio
from ursina import music_system
from ursina.duplicate import duplicate
from panda3d.core import Quat
from ursina.vec2 import Vec2
from ursina.vec3 import Vec3
from ursina.vec4 import Vec4
from ursina.shader import Shader
from ursina.lights import *

from ursina.text import Text
from ursina.mesh import Mesh, MeshModes
from ursina.models.procedural.nine_slice import NineSlice

from ursina.prefabs.sprite import Sprite
from ursina.prefabs.button import Button
from ursina.prefabs.panel import Panel
from ursina.prefabs.sprite_sheet_animation import SpriteSheetAnimation
from ursina.prefabs.animation import Animation
from ursina.prefabs.frame_animation_3d import FrameAnimation3d
from ursina.prefabs.animator import Animator
from ursina.prefabs.sky import Sky
from ursina.prefabs.cursor import Cursor

from ursina.models.procedural.quad import Quad
from ursina.models.procedural.plane import Plane
from ursina.models.procedural.circle import Circle
from ursina.models.procedural.pipe import Pipe
from ursina.models.procedural.cone import Cone
from ursina.models.procedural.cube import Cube
from ursina.models.procedural.cylinder import Cylinder
from ursina.models.procedural.capsule import Capsule
from ursina.models.procedural.grid import Grid
from ursina.models.procedural.terrain import Terrain

from ursina.terraincast import terraincast
from ursina.scripts.smooth_follow import SmoothFollow
from ursina.scripts.grid_layout import grid_layout
from ursina.scripts.scrollable import Scrollable
from ursina.scripts.property_generator import generate_properties_for_class
from ursina.scripts.every_decorator import every

from ursina.prefabs.tooltip import Tooltip
from ursina.prefabs.text_field import TextField
from ursina.prefabs.input_field import InputField, ContentTypes
from ursina.prefabs.draggable import Draggable
from ursina.prefabs.slider import Slider, ThinSlider
from ursina.prefabs.button_group import ButtonGroup
from ursina.prefabs.window_panel import WindowPanel, Space
from ursina.prefabs.button_list import ButtonList
from ursina.prefabs.checkbox import Checkbox
# from ursina.prefabs.file_browser import FileBrowser
# from ursina.prefabs import primitives

# from ursina.prefabs.debug_menu import DebugMenu
from ursina.prefabs.editor_camera import EditorCamera
# from ursina.prefabs.hot_reloader import HotReloader



from ursina import *


app = Ursina()


# player = Entity(model='cube', color=color.orange, scale_y=2)

hit_plane = Entity(model='plane', collider='box', scale=100, alpha=.2, visible=False)

# terrain = Entity(model=Terrain('desert_terrain_heightmap'), scale=128)

cursor = Entity(model='sphere', color=color.azure, scale=1)

# terrain = Mesh()

vertices, triangles = [], []
uvs = []
w, h = 16,16
# self.height_values = [[j/255 for j in i] for i in self.height_values]

centering_offset = Vec2(-.5, -.5)


min_dim = min(w, h)

# from ursina.shaders import lit_with_shadows_shader
terrain = Entity(model=Mesh(vertices=[], triangles=[], uvs=[], colors=[]), scale=(w,1,h), y=-.01,
    # texture='grass',
collider='box')
terrain.scale *= 5

# create the plane
i = 0
for z in range(h):
    for x in range(w):
        terrain.model.vertices.append(Vec3((x/min_dim)+(centering_offset.x), 0, (z/min_dim)+centering_offset.y))
        terrain.model.uvs.append((x/w, z/h))

        if x > 0 and z > 0:
            terrain.model.triangles.append((i, i-1, i-w-1, i-w-0))

        i += 1

# terrain.model.colors = [color.black for v in terrain.model.vertices]
terrain.model.generate()

terrain.model.height_values =[[0 for x in range(w)] for y in range(h)]
# from ursina.prefabs.first_person_controller import FirstPersonController

ec = EditorCamera(rotation_smoothing=0, enabled=1, rotation=(30,30,0))
# player = FirstPersonController()

# def generate_normals_for_heightmap
gradient_colors = {
     '0'  : '#9d9867',
    '11'  : '#9b9563',
    '38'  : '#828131',
    '43'  : '#696327',
    '46'  : '#696327',
    '47'  : '#716d37',
    '49'  : '#6d6828',
    '51'  : '#675f26',
    '54'  : '#5d5b2a',
    '55'  : '#4c483e',
    '58'  : '#645c27',
    '60'  : '#433d2e',
    '62'  : '#5e5a3a',
    '66'  : '#4b473c',
    '73'  : '#3d3425',
    '78'  : '#3a3a39',
    '80'  : '#423d31',
    '82'  : '#2e2d29',
    '88'  : '#3d3c3a',
    '255'  : '#000000',
}
def make_gradient(index_color_dict):    # returns a list of 256 colors
    '''
    given a dict of positions and colors, interpolates the colors into a list of 256 colors
    example input: {'0':color.hex('#9d9867'), '38':color.hex('#828131'), '54':color.hex('#5d5b2a'), '255':color.hex('#000000')}
    '''
    gradient = [color.black for i in range(256)]

    gradient_color_keys = tuple(index_color_dict.keys())
    gradient_color_values = tuple(index_color_dict.values())

    for i in range(len(gradient_colors)):
        from_col = color.hex(gradient_color_values[i-1])
        to_col = color.hex(gradient_color_values[i])

        from_num = 0
        if i > 0:
            from_num = int(gradient_color_keys[i-1])
        to_num = int(gradient_color_keys[i])
        dist = to_num - from_num

        for j in range(dist):
            gradient[from_num+j] = lerp(from_col, to_col, j/dist)

    return gradient

gradient = make_gradient(gradient_colors)


strength = 5
def update():
    if mouse.world_point:
        cursor.position = mouse.world_point

        if mouse.left:
            x = int(cursor.x/(terrain.scale_x/w) + w/2)
            z = int(cursor.z/(terrain.scale_z/h) + h/2)

            heights = []
            for z_offset in range(-3, 3):
                for x_offset in range(-3, 3):
                    true_z, true_x = x+x_offset, z+z_offset
                    if true_x >= 0 and true_x+1 < w and true_z >= 0 and true_z+1 < h:
                        heights.append(terrain.model.height_values[true_x][true_z])
            average_height = sum(heights) / len(heights)
            # print('average:', average_height)
            for z_offset in range(-3, 3):
                for x_offset in range(-3, 3):
                    true_z, true_x = x+x_offset, z+z_offset
                    brush_falloff = 1 - (distance_2d((0,0), (x_offset,z_offset)) / 4)

                    if true_x >= 0 and true_x+1 < w and true_z >= 0 and true_z+1 < h:
                        if not held_keys['shift']:
                            if not held_keys['alt']:
                                terrain.model.height_values[true_z][true_x] += strength * brush_falloff * time.dt
                            else:
                                terrain.model.height_values[true_z][true_x] -= strength * brush_falloff * time.dt
                        else:   #smooth
                            terrain.model.height_values[true_z][true_x] = lerp(terrain.model.height_values[true_z][true_x], average_height, strength * brush_falloff * time.dt)

            terrain.model.vertices = []
            terrain.model.colors = []
            for z, column in enumerate(terrain.model.height_values):
                for x, row in enumerate(column):
                    terrain.model.vertices.append(Vec3(x/w, terrain.model.height_values[x][z], z/h) + Vec3(centering_offset.x, 0, centering_offset.y))
                    # terrain.model.colors.append(hsv(0, 0, 1-(terrain.model.height_values[x][z]*1)))
                    y = int(terrain.model.height_values[x][z]*16)
                    y = clamp(y, 0, 255)
                    terrain.model.colors.append(gradient[y])

            terrain.model.generate()

    pos = cursor.get_position(relative_to=terrain) + Vec3(.5,0,.5)
    if pos.x >= 0 and pos.x < 1 and pos.z >= 0 and pos.z < 1:
        pos *= Vec3(w,0,h)
        # print(int(pos.x), int(pos.z))
        cursor.y = terrain.model.height_values[int(pos.x)][int(pos.z)]
        x, _, z = pos
        # print(floor(x), floor(z))

        height_values = terrain.model.height_values
        point =     height_values[int(floor(x))][int(floor(z))]
        point_e =   height_values[int(min(w-1, ceil(x)))][int(floor(z))]
        point_n =   height_values[int(floor(x))][int(min(h-1, ceil(z)))]
        point_ne =  height_values[int(min(w-1, ceil(x)))][int(min(h-1, ceil(z)))]

        u0v0 = point * (ceil(x) - x) * (ceil(z) - z) # interpolated (x0, z0)
        u1v0 = point_e * (x - floor(x)) * (ceil(z) - z) # interpolated (x1, z0)
        u0v1 = point_n * (ceil(x) - x) * (z - floor(z)) # interpolated (x0, z1)
        u1v1 = point_ne * (x - floor(x)) * (z - floor(z)) # interpolated (x1, z1)

        _h = u0v0 + u1v0 + u0v1 + u1v1  #estimate
        cursor.y = _h * terrain.scale_y



# def input(key):
#     if key == 'tab':    # press tab to toggle edit/play mode
#         ec.enabled = not ec.enabled
#         player.enabled = not player.enabled

Sky()
# DirectionalLight().look_at(Vec3(-.5,-1,-1))
app.run()



from ursina import Shader

camera_noise_shader = Shader(
fragment='''
#version 430

uniform sampler2D tex;
uniform float noise_offset = 1;
uniform float strength = 1;
in vec2 uv;
out vec4 color;

void main() {
    vec3 rgb = texture(tex, uv).rgb;

    float noise = fract(sin(dot(uv+noise_offset, vec2(12.9898, 78.233))) * 43758.5453);
    noise -= .5;
    noise *= strength;

    rgb += noise;

    color = vec4(rgb.r, rgb.g, rgb.b, 1.0);
}

''',
geometry='',
default_input={
    'noise_offset':1,
    'strength':.2
})



if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere', color=color.orange)
    e = Entity(model='cube', y=-1)
    camera.shader = camera_noise_shader
    EditorCamera()

    #camera.set_shader_input("strength", .2)
    Sky()
    def update():
        # Since there isn't really a way to generate pseudorandom numbers in GLSL by themselves, this needs to be put here, otherwise the noise will be static.
        camera.set_shader_input("noise_offset", random.randint(0, 100)/100)

    app.run()



from ursina import *

model_names = ('quad', 'cube', 'sphere', 'plane')

for colorname in color.color_names:
    for modelname in model_names:
        procedural_code = dedent(f'''
            class {colorname.capitalize()}{modelname.capitalize()}(Entity):
                def __init__(self, **kwargs):
                    super().__init__()
                    self.model = '{modelname}'
                    self.color = color.{colorname}
                    for key, value in kwargs.items():
                        setattr(self, key, value)
            ''')
        exec(procedural_code)


if __name__ == '__main__':
    '''procedurally generate classes like RedCube, GreenCube, BlueSphere and so on.'''
    from primitives import *
    app = Ursina()
    RedCube()
    VioletSphere(x=1)
    YellowCube(x=2, scale=(10,1,10), texture='white_cube', texture_scale=(10,10))
    app.run()



from ursina import *




app = Ursina()

window.color = color._16
cursor = Entity(parent=camera.ui, model=Circle(12), color=color.azure, scale=.05, y=.1)
text = Text(parent=cursor, text='x', z=-1, world_scale=25, y=1.25, origin=(0,0))
mesh = Entity(parent=camera.ui, model=Mesh(mode='point', thickness=10), color=color.white).model
# undo_cache = [color.random_color() for e in range(1)]
undo_cache = list()
undo_cache.append(color.random_color())
undo_index = 0
# undo_index = len(undo_cache)-1
help_text = Text(text='undo example\n\nspace: record action\na: undo\nd: redo', origin=(-.5,.5), position=window.top_left*.95)


def visualize_undo_cache():
    global mesh, cursor, undo_index, undo_cache, text
    mesh.vertices = [Vec3(i/10,0,0) for i in range(len(undo_cache))]
    mesh.colors = [e for e in (undo_cache)]
    mesh.generate()
    cursor.x = undo_index / 10
    cursor.color = undo_cache[undo_index]
    text.text = undo_index

visualize_undo_cache()



def input(key):
    global undo_cache, undo_index

    if key == 'a' or key == 'z':
        undo_index -= 1
        undo_index = clamp(undo_index, 0, len(undo_cache)-1)

    if key == 'd' or key == 'y':
        undo_index += 1
        undo_index = clamp(undo_index, 0, len(undo_cache)-1)

    if key == 'space':
        undo_index += 1
        undo_cache = undo_cache[:undo_index]
        undo_cache.append(color.random_color())


    # undo_index = clamp(undo_index, 0, len(undo_cache)-1)
    visualize_undo_cache()


app.run()



import subprocess
import glob
import os
import sys

def main():
    test_files = sorted(glob.glob("*.py"))
    this_script = os.path.basename(__file__)
    for file in test_files:
        if file == this_script:
            continue
        print(f"\n=== Running {file} ===")
        try:
            result = subprocess.run(
                [sys.executable, file],
                check=True
            )
        except subprocess.CalledProcessError as e:
            print(f"\n[ERROR] {file} exited with code {e.returncode}")
            continue
        except KeyboardInterrupt:
            print("\n[INTERRUPTED] Stopping.")
            break

if __name__ == "__main__":
    main()


import math
import os

import psutil

from ursina import Text, Vec2, camera, window


def size(size_bytes):
    if size_bytes == 0:
       return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {size_name[i]}"


class MemoryCounter(Text):
    def __init__(self, **kwargs):
        super().__init__(ignore=False)
        self.parent = camera.ui
        self.position = window.bottom_right - Vec2(.025,0)
        self.origin = (0.5, -0.5)

        self.process = psutil.Process(os.getpid())
        self.i = 0
        self.text = 'eofiwjeofiwejf'

        for key, value in kwargs.items():
            setattr(self, key, value)

    def update(self):
        self.i += 1
        if self.i > 10:
            self.text = str(size(self.process.memory_info().rss))

            self.i = 0


if __name__ == '__main__':
    from ursina import Ursina
    app = Ursina()
    MemoryCounter()
    '''
    Displays the amount of memory used in the bottom right corner
    '''
    app.run()



from ursina import *


def combine(combine_parent, analyze=False, auto_destroy=True, ignore=[], ignore_disabled=True, include_normals=False):
    if not combine_parent.children:
        print_warning('Error, trying to combine children of entity with no children.', combine_parent.children)
        return

    verts = []
    tris = []
    norms = []
    uvs = []
    cols = []
    to_destroy = []
    o = 0
    original_texture = combine_parent.texture

    for e in scene.entities:
        if e in ignore:
            continue

        if ignore_disabled and not e.enabled:
            continue

        if e.has_ancestor(combine_parent) or e == combine_parent:
            if not hasattr(e, 'model') or e.model == None or e.scripts or e.eternal:
                continue
            if not hasattr(e.model, 'vertices') or not e.model.vertices:
                e.model = load_model(e.model.name, use_deepcopy=True)
                e.origin = e.origin
            if not e.model:
                continue

            if analyze:
                print('combining:', e)

            vertex_to_world_matrix = e.model.getTransform(combine_parent).getMat()
            verts += [Vec3(*vertex_to_world_matrix.xformPoint(Vec3(*v))) for v in e.model.vertices]

            if not e.model.triangles:
                new_tris = [i for i in range(len(e.model.vertices))]

            else:
                new_tris = list()
                for t in e.model.triangles:
                    if isinstance(t, int):
                        new_tris.append(t)
                    elif len(t) == 3:
                        new_tris.extend(t)
                    elif len(t) == 4: # turn quad into tris
                        new_tris.extend([t[0], t[1], t[2], t[2], t[3], t[0]])

            new_tris = [t+o for t in new_tris]
            new_tris = [(new_tris[i], new_tris[i+1], new_tris[i+2]) for i in range(0, len(new_tris)-1, 3)]

            o += len(e.model.vertices)
            tris += new_tris

            if e.model.uvs:
                uvs.extend([(uv * e.texture_scale) + e.texture_offset for uv in e.model.uvs])
            else:
                uvs.extend([(0,0) for e in e.model.vertices])

            if e.model.colors: # if has vertex colors
                cols.extend([Color(*vcol) * e.color for vcol in e.model.colors])
            else:
                cols.extend((e.color, ) * len(e.model.vertices))

            if include_normals:
                rotation_scale_matrix = vertex_to_world_matrix.getUpper3()  # get the rotation and scaling part of the matrix
                normal_to_world_matrix = rotation_scale_matrix
                normal_to_world_matrix.invertInPlace()
                normal_to_world_matrix.transposeInPlace()

                if e.model.normals: # if has normals
                    norms.extend([Vec3(*normal_to_world_matrix.xform(Vec3(*n)))for n in e.model.normals])
                else:
                    norms.extend((Vec3.up, ) * len(e.model.vertices))


            if auto_destroy and e != combine_parent:
                to_destroy.append(e)

    if auto_destroy:
        from ursina import destroy
        [destroy(e) for e in to_destroy]

    combine_parent.model = Mesh(vertices=verts, triangles=tris, normals=norms, uvs=uvs, colors=cols, mode='triangle')
    combine_parent.texture = original_texture
    # print('combined')
    # entity.model = Mesh(vertices=verts,  mode='triangle')
    # entity.flatten_strong()
    if analyze:
        render.analyze()
    return combine_parent.model




if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    from ursina.shaders import lit_with_shadows_shader, normals_shader
    Entity.default_shader = lit_with_shadows_shader

    combine_parent = Entity(texture='brick')
    e1 = Entity(parent=combine_parent, model='sphere', y=1.5, color=color.pink, rotation_y=90)
    e2 = Entity(parent=combine_parent, model='cube', color=color.yellow, x=1, origin_y=-.5, texture='brick')
    e3 = Entity(parent=e2, model='cube', color=color.yellow, y=2, scale=.5, texture='brick', texture_scale=Vec2(3,3), texture_offset=(.1,.1))
    e4 = Entity(parent=combine_parent, model='plane', color=color.lime, scale=10, texture='brick', texture_scale=Vec2(5,5))
    Text(position=window.top, origin=(0,.5), text='Press space to combine')

    def input(key):
        if key == 'space':
            from time import perf_counter
            t = perf_counter()
            combine_parent.combine(include_normals=True)
            combine_parent.shader = lit_with_shadows_shader
            # combine_parent.shader = normals_shader
            print('combined in:', perf_counter() - t)


    from ursina.lights import DirectionalLight
    sun = DirectionalLight()
    sun.look_at(Vec3(-1,-1,-1))

    EditorCamera()
    app.run()



from ursina import *
from ursina.shaders import lit_with_shadows_shader


app = Ursina(size=(1280,720))

physics_entities = []
class PhysicsEntity(Entity):
    def __init__(self, model='cube', collider='box', **kwargs):
        super().__init__(model=model, collider=collider, **kwargs)
        physics_entities.append(self)

    def update(self):
        if self.intersects():
            self.stop()
            return

        self.velocity = lerp(self.velocity, Vec3(0), time.dt)
        self.velocity += Vec3(0,-1,0) * time.dt * 5
        self.position += (self.velocity + Vec3(0,-4,0)) * time.dt

    def stop(self):
        self.velocity = Vec3(0,0,0)
        if self in physics_entities:
            physics_entities.remove(self)

    def on_destroy(self):
        self.stop()


    def throw(self, direction, force):
        pass


Entity.default_shader = lit_with_shadows_shader

ground = Entity(model='plane', scale=32, texture='white_cube', texture_scale=Vec2(32), collider='box', color=color.light_gray)
DirectionalLight().look_at(Vec3(1,-1,-1))

from ursina.prefabs.first_person_controller import FirstPersonController
player = FirstPersonController()
Sky()


def input(key):
    if key == 'left mouse down':
        e = PhysicsEntity(model='cube', color=color.azure, velocity=Vec3(0), position=player.position+Vec3(0,1.5,0)+player.forward, collider='sphere')
        e.velocity = (camera.forward + Vec3(0,.5,0)) * 10
        # physics_entities.append(e)


app.run()



from ursina import *

if __name__ == '__main__':
    app = Ursina()

camera.orthographic = True
camera.fov = 4
camera.position = (1, 1)
Text.default_resolution *= 2

player = Entity(name='o', color=color.azure)
cursor = Tooltip(player.name, color=player.color, origin=(0,0), scale=4, enabled=True)
cursor.background.color = color.clear
bg = Entity(parent=scene, model='quad', texture='shore', scale=(16,8), z=10, color=color.light_gray)
mouse.visible = False

# create a matrix to store the buttons in. makes it easier to check for victory
board = [[None for x in range(3)] for y in range(3)]

for y in range(3):
    for x in range(3):
        b = Button(parent=scene, position=(x,y))
        board[x][y] = b

        def on_click(b=b):
            b.text = player.name
            b.color = player.color
            b.collision = False
            check_for_victory()

            if player.name == 'o':
                player.name = 'x'
                player.color = color.orange
            else:
                player.name = 'o'
                player.color = color.azure

            cursor.text = player.name
            cursor.color = player.color

        b.on_click = on_click


def check_for_victory():
    name = player.name

    won = (
    (board[0][0].text == name and board[1][0].text == name and board[2][0].text == name) or # across the bottom
    (board[0][1].text == name and board[1][1].text == name and board[2][1].text == name) or # across the middle
    (board[0][2].text == name and board[1][2].text == name and board[2][2].text == name) or # across the top
    (board[0][0].text == name and board[0][1].text == name and board[0][2].text == name) or # down the left side
    (board[1][0].text == name and board[1][1].text == name and board[1][2].text == name) or # down the middle
    (board[2][0].text == name and board[2][1].text == name and board[2][2].text == name) or # down the right side
    (board[0][0].text == name and board[1][1].text == name and board[2][2].text == name) or # diagonal /
    (board[0][2].text == name and board[1][1].text == name and board[2][0].text == name))   # diagonal \

    if won:
        print('winner is:', name)
        cursor.text = ''
        mouse.visible = True
        Panel(z=1, scale=10, model='quad')
        t = Text(f'player\n{name}\nwon!', scale=3, origin=(0,0), background=True)
        t.create_background(padding=(.5,.25), radius=Text.size/2)
        t.background.color = player.color.tint(-.2)

if __name__ == '__main__':
    app.run()



from pathlib import Path
from enum import Enum
from textwrap import dedent
import numbers
import array

from ursina import application
from ursina import color
from ursina.color import Color
from ursina.scripts.generate_normals import generate_normals
from ursina.scripts.project_uvs import project_uvs
from ursina.scripts.colorize import colorize
from ursina.array_tools import LoopingList
from ursina.vec3 import Vec3
from ursina.vec2 import Vec2
from ursina.sequence import Func

import panda3d.core as p3d

try:
    from warnings import deprecated
except:
    from ursina.scripts.deprecated_decorator import deprecated


class MeshModes(Enum):
    triangle = 'triangle'
    ngon = 'ngon'
    quad = 'quad'
    line = 'line'
    point = 'point'
    tristrip = 'tristrip'

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other): # overridden __eq__ to allow for both str and MeshModes comparisons
        if isinstance(other, MeshModes):
            return self.value == other.value
        return self.value == other



class Mesh(p3d.NodePath):
    _modes = {
        'triangle' : p3d.GeomTriangles,
        'tristrip' : p3d.GeomTristrips,
        'ngon' : p3d.GeomTrifans,
        'line' : p3d.GeomLinestrips,
        'point' : p3d.GeomPoints,
    }

    def __init__(self, vertices=None, triangles=None, colors=None, uvs=None, normals=None, static=True, mode='triangle', thickness=1, render_points_in_3d=True, vertex_buffer=None, vertex_buffer_length=None, vertex_buffer_format=None):
        super().__init__('mesh')
        self.vertices = vertices
        self.triangles = triangles
        self.colors = colors
        self.uvs = uvs
        self.normals = normals
        self.static = static
        self.mode = mode
        self.thickness = thickness
        self.render_points_in_3d = render_points_in_3d
        self.vertex_buffer = vertex_buffer
        self.vertex_buffer_length = vertex_buffer_length
        self.vertex_buffer_format = vertex_buffer_format

        self._generated_vertices = None

        for var in (('vertices', vertices), ('triangles', triangles), ('colors', colors), ('uvs', uvs), ('normals', normals)):
            name, value = var
            if value is None:
                setattr(self, name, [])

        if (self.vertices is not None and len(self.vertices) > 0) or self.vertex_buffer is not None:
            self.generate()


    def _ravel(self, data):
        if not isinstance(data[0], numbers.Real):
            d = []
            for v in data:
                d.extend(v)
            return d
        return data


    def _set_array_data(self, array_handle, data, dtype_string='f'):
        a = None
        try:
            a = memoryview(data).cast('B').cast(dtype_string)
        except:
            a = array.array(dtype_string, data)

        vmem = memoryview(array_handle).cast('B').cast(dtype_string)
        try:
            vmem[:] = a
        except:
            raise Exception(f'Error in Mesh. Ensure Mesh is valid and the inputs have same length: vertices:{len(self.vertices)}, triangles:{len(self.triangles)}, normals:{len(self.normals)}, colors:{len(self.colors)}, uvs:{len(self.uvs)}')

    def generate(self): # Must be called after a mesh's values has been updated in order to update visually
        self._generated_vertices = None

        if hasattr(self, 'geomNode'):
            self.geomNode.removeAllGeoms()
        else:
            self.geomNode = p3d.GeomNode('mesh_geom')
            self.attachNewNode(self.geomNode)

        if len(self.vertices) == 0 and self.vertex_buffer is None:
            return

        static_mode = p3d.Geom.UHStatic if self.static else p3d.Geom.UHDynamic
        vertex_format = p3d.GeomVertexFormat()

        color_attribute_index = -1
        uv_attribute_index = -1
        normal_attribute_index = -1

        if self.vertex_buffer is not None:
            vertex_array_format = p3d.GeomVertexArrayFormat()
            attributes = self.vertex_buffer_format.split(",")
            for attribute in attributes:
                attribute_type = attribute[0]
                attribute_type_name = attribute[0]
                if attribute_type == "p":
                    attribute_type = p3d.Geom.C_point
                    attribute_type_name = "vertex"
                elif attribute_type == "c":
                    attribute_type = p3d.Geom.C_color
                    attribute_type_name = "color"
                elif attribute_type == "t":
                    attribute_type = p3d.Geom.C_texcoord
                    attribute_type_name = "texcoord"
                elif attribute_type == "n":
                    attribute_type = p3d.Geom.C_normal
                    attribute_type_name = "normal"
                else:
                    raise Exception("Invalid vertex buffer format attribute type: {}.".format(attribute_type))
                attribute_count = int(attribute[1:-1])
                attribute_dtype = attribute[-1]
                if attribute_dtype == "f":
                    attribute_dtype = p3d.Geom.NT_float32
                elif attribute_dtype == "I":
                    attribute_dtype = p3d.Geom.NT_uint32
                else:
                    raise Exception("Invalid vertex buffer format attribute data type: {}.".format(attribute_dtype))
                vertex_array_format.addColumn(attribute_type_name, attribute_count, attribute_dtype, attribute_type)
            vertex_format.addArray(vertex_array_format)

        else:
            vertex_format = p3d.GeomVertexFormat()
            vertex_format.add_array(p3d.GeomVertexFormat.getV3().arrays[0])
            attribute_count = 1
            if self.colors is not None and len(self.colors) > 0:
                vertex_format.add_array(p3d.GeomVertexArrayFormat('color', 4, p3d.Geom.NT_float32, p3d.Geom.C_color))
                color_attribute_index = attribute_count
                attribute_count += 1
            if self.uvs is not None and len(self.uvs) > 0 and self.mode not in ['line', 'point']:
                vertex_format.add_array(p3d.GeomVertexArrayFormat('texcoord', 2, p3d.Geom.NT_float32, p3d.Geom.C_texcoord))
                uv_attribute_index = attribute_count
                attribute_count += 1
            if self.normals is not None and len(self.normals) > 0 and self.mode not in ['line', 'point']:
                vertex_format.add_array(p3d.GeomVertexArrayFormat('normal', 3, p3d.Geom.NT_float32, p3d.Geom.C_normal))
                normal_attribute_index = attribute_count
                attribute_count += 1

        vertex_format = p3d.GeomVertexFormat.register_format(vertex_format)
        vdata = p3d.GeomVertexData('vertex_data', vertex_format, static_mode)

        if self.vertex_buffer is not None:
            m = memoryview(self.vertex_buffer).cast('B')
            vdata.unclean_set_num_rows(self.vertex_buffer_length)
            array_handle = vdata.modify_array(0)
            vmem = memoryview(array_handle).cast('B')
            vmem[:] = m

        else:
            if isinstance(self.vertices[0], numbers.Real):
                vdata.unclean_set_num_rows(len(self.vertices) // 3)
            else:
                vdata.unclean_set_num_rows(len(self.vertices))

            self._set_array_data(vdata.modify_array(0), self._ravel(self.vertices), 'f')

            if self.colors is not None and len(self.colors) > 0:
                self._set_array_data(vdata.modify_array(color_attribute_index), self._ravel(self.colors), 'f')

            if self.uvs is not None and len(self.uvs) > 0 and self.mode not in ['line', 'point']:
                self._set_array_data(vdata.modify_array(uv_attribute_index), self._ravel(self.uvs), 'f')

            if self.normals is not None and len(self.normals) > 0 and self.mode not in ['line', 'point']:
                self._set_array_data(vdata.modify_array(normal_attribute_index), self._ravel(self.normals), 'f')

        geom = p3d.Geom(vdata)
        self.geom = geom

        if len(self.triangles) == 0:    # no triangles provided, so just add them in order
            prim = Mesh._modes[self.mode](static_mode)
            prim.set_index_type(p3d.GeomEnums.NT_uint32)

            parray = prim.modify_vertices()
            n = len(self.vertices)
            if isinstance(self.vertices[0], numbers.Real):
                n = n // 3
            triangles = [i for i in range(n)]
            parray.unclean_set_num_rows(n)
            self._set_array_data(parray, triangles, 'I')

            prim.close_primitive()
            geom.addPrimitive(prim)

        else:
            if not isinstance(self.triangles[0], numbers.Real): # triangles provided as [(0,1,2), (3,4,5,6), ...] etc., so unpack them
                line_segments = []
                indices = self.indices
                for tup in self.triangles:
                    if len(tup) == 2:
                        line_segments.append(tup)


                for line_segment in line_segments:
                    prim = Mesh._modes['line'](static_mode)
                    prim.set_index_type(p3d.GeomEnums.NT_uint32)
                    parray = prim.modify_vertices()
                    parray.unclean_set_num_rows(len(line_segment))
                    self._set_array_data(parray, line_segment, 'I')
                    prim.close_primitive()
                    geom.addPrimitive(prim)

                if len(indices) > 0:
                    prim = Mesh._modes[self.mode](static_mode)
                    prim.set_index_type(p3d.GeomEnums.NT_uint32)
                    parray = prim.modify_vertices()
                    parray.unclean_set_num_rows(len(indices))
                    self._set_array_data(parray, indices, 'I')
                    prim.close_primitive()
                    geom.addPrimitive(prim)

            else:   # got triangles as [0,1,2,3,4,5], ie. flat
                prim = Mesh._modes[self.mode](static_mode)
                prim.set_index_type(p3d.GeomEnums.NT_uint32)

                parray = prim.modify_vertices()

                parray.unclean_set_num_rows(len(self.triangles))
                self._set_array_data(parray, self._ravel(self.triangles), 'I')

                prim.close_primitive()
                geom.addPrimitive(prim)

        self.geomNode.addGeom(geom)

        if self.mode == 'point':
            self.setTexGen(p3d.TextureStage.getDefault(), p3d.TexGenAttrib.MPointSprite)
            # self.setShaderAuto()


    @property
    def indices(self):  # Get the vertex indices as a flat list. For example if you have tuples in triangles :((0,1,2),(3,4,5)) -> (0,1,2,3,4,5). Or with quads: ((0,1,2,3),) -> (0,1,2,2,3,0).
        if not self.triangles:
            return list(range(len(self.vertices)))

        if self.triangles and isinstance(self.triangles[0], numbers.Real):
            return self.triangles

        indices = []
        for tup in self.triangles:
            # if len(tup) == 2:
            #     line_segments.append(tup)
            if len(tup) == 3:
                indices.extend(tup)
            elif len(tup) == 4:
                indices.extend((tup[0], tup[1], tup[2],
                                  tup[2], tup[3], tup[0]))
            elif len(tup) > 4:
                tup = LoopingList(tup)
                for i in range(1, len(tup)):
                    indices.extend((tup[0], tup[i], tup[i+1]))

        return indices


    @property
    def generated_vertices(self):
        if self._generated_vertices is None:
            if self.triangles is not None and len(self.triangles) > 0:
                if not isinstance(self.triangles[0], numbers.Real):
                    tris = []
                    for tup in self.triangles:
                        if len(tup) == 4:
                            tris.extend((tup[0], tup[1], tup[2],
                                         tup[2], tup[3], tup[0]))
                        else:
                            tris.extend(tup)
                    self._generated_vertices = [self.vertices[i] for i in tris]
                else:
                    self._generated_vertices = [self.vertices[i] for i in self.triangles]
            else:
                self._generated_vertices = self.vertices
        return self._generated_vertices

    @generated_vertices.setter
    def generated_vertices(self, value):
        self._generated_vertices = value


    def serialize(self, vertex_decimal_limit=4, color_decimal_limit=4, uv_decimal_limit=4, normal_decimal_limit=4):
        vbuf_format = self.vertex_buffer_format
        if vbuf_format is not None:
            vbuf_format = f'"{vbuf_format}"'

        mesh_as_string = 'Mesh('
        mesh_as_string += f'\n    vertices={[tuple(round(e, vertex_decimal_limit) for e in vert) for vert in self.vertices]},' if self.vertices else ''
        mesh_as_string += f'\n    triangles={self.triangles},' if self.triangles else ''
        mesh_as_string += f'\n    colors={[tuple(round(e, color_decimal_limit) for e in col) for col in self.colors]},' if self.colors else ''
        mesh_as_string += f'\n    uvs={[tuple(round(e, uv_decimal_limit) for e in uv) for uv in self.uvs]},' if self.uvs else ''
        mesh_as_string += f'\n    normals={[tuple(round(e, normal_decimal_limit) for e in norm) for norm in self.normals]},' if self.normals else ''
        mesh_as_string += f'\n    static={self.static},' if not self.static else ''
        mesh_as_string += f'\n    mode="{self.mode}",' if self.mode != 'triangle' else ''
        mesh_as_string += f'\n    thickness={self.thickness},' if self.thickness != 1 else ''
        mesh_as_string += f'\n    render_points_in_3d={self.render_points_in_3d},' if not self.render_points_in_3d else ''
        mesh_as_string += f'\n    vertex_buffer={self.vertex_buffer},' if self.vertex_buffer is not None else ''
        mesh_as_string += f'\n    vertex_buffer_length={self.vertex_buffer_length},' if self.vertex_buffer_length is not None else ''
        mesh_as_string += f'\n    vertex_buffer_format={vbuf_format},' if self.vertex_buffer_format is not None else ''

        mesh_as_string += '\n    )'
        return mesh_as_string


    @property
    def render_points_in_3d(self):
        return self._render_points_in_3d

    @render_points_in_3d.setter
    def render_points_in_3d(self, value):
        self._render_points_in_3d = value
        self.set_render_mode_perspective(value)

    def __repr__(self):
        if not self.name == 'mesh':
            return self.name
        else:
            return self.serialize()

    def __str__(self):
        if hasattr(self, 'name'):
            return self.name

    def __add__(self, other):
        if self.vertex_buffer is not None:
            raise Exception("Can't add mesh with vertex buffer to another mesh (operation not supported).")

        self.vertices += other.vertices
        self.triangles += other.triangles
        if other.colors:
            self.colors += other.colors
        else:
            self.colors += (color.white, ) * len(other.vertices)
        self.normals += other.normals
        self.uvs += other.uvs

    def __deepcopy__(self, memo):
        m = Mesh(
            vertices=[Vec3(*e) for e in self.vertices],
            triangles=self.triangles,
            colors=[Color(*e) for e in self.colors],
            uvs=[Vec2(*e) for e in self.uvs],
            normals=[Vec3(*e) for e in self.normals],
            static=self.static,
            mode=self.mode,
            thickness=self.thickness,
            render_points_in_3d=self.render_points_in_3d,
            vertex_buffer=self.vertex_buffer,
            vertex_buffer_length=self.vertex_buffer_length,
            vertex_buffer_format=self.vertex_buffer_format
        )
        m.name = self.name
        return m

    @property
    def thickness(self):
        return self.getRenderModeThickness()

    @thickness.setter
    def thickness(self, value):
        self.setRenderModeThickness(value)

    def generate_normals(self, smooth=True, regenerate=True):
        self.normals = list(generate_normals(self.vertices, self.indices, smooth))
        if regenerate:
            self.generate()
        return self.normals

    def colorize(self, left=color.white, right=color.blue, down=color.red, up=color.green, back=color.white, forward=color.white, smooth=True, world_space=True, strength=1):
        colorize(self, left, right, down, up, back, forward, smooth, world_space, strength)

    def project_uvs(self, aspect_ratio=1, direction=Vec3.forward):
        project_uvs(self, aspect_ratio, direction=direction)

    def clear(self, regenerate=True):
        if self.vertex_buffer is not None:
            self.vertex_buffer = None
            self.vertex_buffer_length = None
            self.vertex_buffer_format = None
        self.vertices = []
        self.triangles = []
        self.colors = []
        self.uvs = []
        self.normals = []
        if regenerate:
            self.generate()

    def save(self, name='', folder:Path=Func(getattr, application, 'models_compressed_folder'), flip_faces=False, vertex_decimal_limit=5, color_decimal_limit=4):
        if callable(folder):
            folder = folder()
        if not folder.exists():
            folder.mkdir()

        if not name and hasattr(self, 'path'):
            name = self.path.stem
        if '.' not in name:
            name += '.ursinamesh'

        if name.endswith('ursinamesh'):
            with open(folder / name, 'w') as f:
                f.write(self.serialize(vertex_decimal_limit=vertex_decimal_limit, color_decimal_limit=color_decimal_limit))
            print('saved .ursinamesh to:', folder / name, 'vertex_decimal_limit:', vertex_decimal_limit, 'color_decimal_limit:', color_decimal_limit)


        elif name.endswith('.obj'):
            from ursina.mesh_exporter import ursinamesh_to_obj
            import os
            name = str(os.path.splitext(name)[0])
            ursinamesh_to_obj(self, name, folder, flip_faces=flip_faces)
        elif name.endswith('.dae'):
            from ursina.mesh_exporter import ursinamesh_to_dae
            import os
            name = str(os.path.splitext(name)[0])
            ursinamesh_to_dae(self, name, folder)
        elif name.endswith('.bam'):
            self.writeBamFile(folder / name)
            print('saved .bam to:', folder / name)


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    from ursina.shaders.unlit_shader import unlit_shader
    Entity.default_shader = unlit_shader

    # verts as list of tuples
    e = Entity(position=(0,0), model=Mesh(vertices=[(-.5,0,0), (.5,0,0), (0, 1, 0)]))

    # verts as tuple of tuples
    e = Entity(position=(1,0), model=Mesh(vertices=((-.5,0,0), (.5,0,0), (0, 1, 0))))
    Text(parent=e, text='triangle mesh\nwith verts as tuple of tuples', y=1, scale=5, origin=(0,-.5))

    # verts as list of lists
    e = Entity(position=(0,-2), model=Mesh(vertices=[[-.5,0,0], [.5,0,0], [0, 1, 0]]))
    Text(parent=e, text='triangle mesh\nwith verts as list of lists', y=1, scale=5, origin=(0,-.5))

    # verts as tuple of lists
    e = Entity(position=(1,-2), model=Mesh(
        vertices=([-.5,0,0], [.5,0,0], [0, 1, 0])
    ))
    Text(parent=e, text='triangle mesh\nwith verts as tuple of lists', y=1, scale=5, origin=(0,-.5))

    # verts as list Vec3
    e = Entity(position=(0,-4), model=Mesh(
        vertices=[Vec3(-.5,0,0), Vec3(.5,0,0), Vec3(0, 1, 0)],
    ))
    Text(parent=e, text='triangle mesh\nwith verts as list Vec3', y=1, scale=5, origin=(0,-.5))


    # tris as flat list
    e = Entity(position=(1,-4), model=Mesh(
        vertices=[Vec3(-.5,0,0), Vec3(.5,0,0), Vec3(0, 1, 0)],
        triangles = [0,1,2],
    ))
    Text(parent=e, text='triangle mesh\nwith tris as flat list', y=1, scale=5, origin=(0,-.5))

    # tris as list of triangles
    e = Entity(position=(2.5,0), model=Mesh(
        vertices=[Vec3(-.5,0,0), Vec3(.5,0,0), Vec3(0, 1, 0)],
        triangles = [(0,1,2), (2,1,0)], # should be double sided
    ))
    Text(parent=e, text='triangle mesh\nwith tris as list of triangles', y=1, scale=5, origin=(0,-.5))


    continious_line = Entity(position=(4,0), model=Mesh(
        vertices=(Vec3(0,0,0), Vec3(.6,.3,0), Vec3(1,1,0), Vec3(.6,1.7,0), Vec3(0,2,0)),
        # triangles= ((0,1), (3,4,5)),
        mode='line',
        thickness=4,
        ), color=color.cyan)
    Text(parent=continious_line, text='continious_line', y=1, scale=5)

    line_segments = Entity(position=(4,-2), model=Mesh(
        vertices=(Vec3(0,0,0), Vec3(.6,.3,0), Vec3(1,1,0), Vec3(.6,1.7,0), Vec3(0,2,0)),
        triangles= ((0,1), (3,4)),
        mode='line',
        thickness=4,
        ), color=color.magenta)
    Text(parent=line_segments, text='line_segments', y=1, scale=5)

    points_3d = Entity(position=(6,0), model=Mesh(vertices=(Vec3(0,0,0), Vec3(.6,.3,0), Vec3(1,1,0), Vec3(.6,1.7,0), Vec3(0,2,0)), mode='point', thickness=.05),
        color=color.red,
        texture='circle',
        )
    # points_3d.setShaderAuto()
    # points_3d.model.setShaderAuto()
    # points_3d.shader = None
    Text(parent=points_3d, text='points_3d', y=1, scale=5)

    points_2d = Entity(position=(6,-2), model=Mesh(vertices=(Vec3(0,0,0), Vec3(.6,.3,0), Vec3(1,1,0), Vec3(.6,1.7,0), Vec3(0,2,0)), mode='point', thickness=10, render_points_in_3d=False), color=color.red, texture='circle')
    Text(parent=points_2d, text='points_2d', y=1, scale=5)
    points_2d.set_shader_auto()
    print('----------------------', points_2d.model.getShader())

    quad = Entity(
        position=(8,0),
        model=Mesh(
            vertices=((0.5, 0.5, 0.0), (-0.5, 0.5, 0.0), (-0.5, -0.5, 0.0), (0.5, -0.5, 0.0), (0.5, 0.5, 0.0), (-0.5, -0.5, 0.0)),
            uvs=((1, 1), (0, 1), (0, 0), (1, 0), (1, 1), (0, 0)),
            mode='triangle'),
        texture='shore'
        )
    Text(parent=quad, text='quad_with_uvs', y=1, scale=5, origin=(0,-.5))

    quad = Entity(
        position=(8,-2),
        model=Mesh(
            vertices=((0.5, 0.5, 0.0), (-0.5, 0.5, 0.0), (-0.5, -0.5, 0.0), (0.5, -0.5, 0.0), (0.5, 0.5, 0.0), (-0.5, -0.5, 0.0)),
            uvs=((1, 1), (0, 1), (0, 0), (1, 0), (1, 1), (0, 0)),
            normals=[(-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0)],
            mode='triangle'),
        )
    Text(parent=quad, text='quad_with_normals', y=1, scale=5, origin=(0,-.5))

    quad = Entity(
        position=(8,-4),
        model=Mesh(
            vertices=((0.5, 0.5, 0.0), (-0.5, 0.5, 0.0), (-0.5, -0.5, 0.0), (0.5, -0.5, 0.0), (0.5, 0.5, 0.0), (-0.5, -0.5, 0.0)),
            uvs=((1, 1), (0, 1), (0, 0), (1, 0), (1, 1), (0, 0)),
            normals=[(-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0)],
            mode='triangle'),
        texture='shore',
        )
    Text(parent=quad, text='quad_with_usv_and_normals', y=1, scale=5, origin=(0,-.5))

    quad = Entity(
        position=(8,-6),
        model=Mesh(
            vertices=((0.5, 0.5, 0.0), (-0.5, 0.5, 0.0), (-0.5, -0.5, 0.0), (0.5, -0.5, 0.0), (0.5, 0.5, 0.0), (-0.5, -0.5, 0.0)),
            uvs=((1, 1), (0, 1), (0, 0), (1, 0), (1, 1), (0, 0)),
            normals=[(-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0), (-0.0, 0.0, -1.0)],
            colors=[color.red, color.yellow, color.green, color.cyan, color.blue, color.magenta],
            mode='triangle'),
        texture='shore'
        )
    Text(parent=quad, text='quad_with_usv_and_normals_and_vertex_colors', y=1, scale=5, origin=(0,-.5))

    quad = Entity(
        position=(10,0),
        model=Mesh(
            vertices=((-0.5, -0.5, 0.0), (0.5, -0.5, 0.0), (0.5, 0.5, 0.0), (-0.5, 0.5, 0.0)),
            triangles=(0,1,2, 2,3,0),
            mode='triangle'),
        )
    Text(parent=quad, text='triangles flat', y=1, scale=5, origin=(0,-.5))

    quad = Entity(
        position=(10,-2),
        model=Mesh(
            vertices=((-0.5, -0.5, 0.0), (0.5, -0.5, 0.0), (0.5, 0.5, 0.0), (-0.5, 0.5, 0.0)),
            triangles=((0,1,2), (2,3,0)),
            mode='triangle'),
        )
    Text(parent=quad, text='triangles triplets', y=1, scale=5, origin=(0,-.5))

    quad = Entity(
        position=(10,-4),
        model=Mesh(
            vertices=((-0.5, -0.5, 0.0), (0.5, -0.5, 0.0), (0.5, 0.5, 0.0), (-0.5, 0.5, 0.0)),
            triangles=((0,1,2,3), (0,3,2)),
            mode='triangle'),
        )
    Text(parent=quad, text='triangles quad + tri', y=1, scale=5, origin=(0,-.5))

    copy_test = Entity(position=(12,0), model=copy(quad.model))
    Text(parent=copy_test, text='copy_test', y=1, scale=5, origin=(0,-.5))

    deepcopy_test = Entity(position=(12,-2), model=deepcopy(quad.model))
    Text(parent=deepcopy_test, text='deepcopy_test', y=1, scale=5, origin=(0,-.5))

    clear_test = Entity(position=(12,-4), model=deepcopy(quad.model))
    clear_test.model.clear()
    Text(parent=clear_test, text='.clear() test', y=1, scale=5, origin=(0,-.5))

    # recipe_test = Entity(position=(12,-6), model=eval(quad.model.recipe))
    # Text(parent=recipe_test, text='.recipe test', y=1, scale=5, origin=(0,-.5))

    window.color = color.black
    EditorCamera()

    # m = Mesh()
    # print(load_model('sphere', application.internal_models_compressed_folder, use_deepcopy=True).serialize())
    # print()
    app.run()



from ursina.shaders.unlit_shader import unlit_shader
from ursina.shaders.lit_with_shadows_shader import lit_with_shadows_shader
from ursina.shaders.normals_shader import normals_shader
from ursina.shaders.basic_lighting_shader import basic_lighting_shader
from ursina.shaders.triplanar_shader import triplanar_shader
from ursina.shaders.matcap_shader import matcap_shader
from ursina.shaders.colored_lights_shader import colored_lights_shader
from ursina.shaders.transition_shader import transition_shader



# post processing effects
from ursina.shaders.screenspace_shaders.camera_vertical_blur import camera_vertical_blur_shader
from ursina.shaders.screenspace_shaders.camera_contrast import camera_contrast_shader
from ursina.shaders.screenspace_shaders.camera_grayscale import camera_grayscale_shader
from ursina.shaders.screenspace_shaders.fxaa import fxaa_shader
from ursina.shaders.screenspace_shaders.ssao import ssao_shader



from ursina import *
from ursina.scripts.property_generator import generate_properties_for_class


class ContentTypes:
    int = '0123456789'
    float = int + '.,'
    int_math = int + '+-*/'
    math = float + '+-*/'

@generate_properties_for_class()
class InputField(Button):
    def __init__(self, default_value='', label='', max_lines=1, character_limit=24, text='', active=False, **kwargs):
        if not 'scale' in kwargs and not 'scale_y' in kwargs:
            kwargs['scale'] = (kwargs.get('scale_x', .5), Text.size*2*max_lines)


        super().__init__(highlight_scale=1, pressed_scale=1, highlight_color=color.black, **kwargs)
        for key, value in kwargs.items():
            if 'scale' in key:
                setattr(self, key, value)

        self.default_value = default_value
        self.limit_content_to = None
        self.hide_content = False   # if set to True, will display content as '*'. can also be set to character instead of True.

        self.next_field = None
        self.submit_on = []   # for example: self.submit_on = ['enter', ] will call self.on_submit when you press enter.
        self.on_submit = None   # function to be called when you press self.submit_on.
        self.on_value_changed = None

        self.text_field = TextField(world_parent=self, x=-.45, y=.25, z=-.1, max_lines=max_lines, character_limit=character_limit, text=text, register_mouse_input=True, highlight_color=hsv(200,1,1,.5))
        destroy(self.text_field.bg)
        self.text_field.bg = self

        def render():
            if self.limit_content_to:
                org_length = len(self.text_field.text)
                self.text_field.text = ''.join([e for e in self.text_field.text if e in self.limit_content_to])
                self.text_field.cursor.x -= org_length - len(self.text_field.text)
                self.text_field.cursor.x = max(0, self.text_field.cursor.x)

            if self.hide_content:
                replacement_char = '*'
                if isinstance(self.hide_content, str):
                    replacement_char = self.hide_content

                self.text_field.text_entity.text = replacement_char * len(self.text_field.text)
                return

            if self.on_value_changed and not self.text_field.text_entity.text == self.text_field.text:
                self.on_value_changed()
            self.text_field.text_entity.text = self.text_field.text

        self.text_field.render = render

        self.text_field.scale *= 1.25
        self.text_field.text = default_value
        self.text_field.render()
        self.text_field.shortcuts['indent'] = ('')
        self.text_field.shortcuts['dedent'] = ('')

        self.active = active

        if label:
            self.label = Text(str(label) + ':', parent = self, position = self.text_field.position, scale = 1.25)
            self.text_field.x += 0.1 * (len(str(label)) + 1.0) / 6.0

        for key, value in kwargs.items():
            setattr(self, key, value)


    def input(self, key):
        if key == 'tab' and self.text_field.cursor.y >= self.text_field.max_lines-1 and self.active:
            self.active = False
            if self.next_field:
                mouse.position = self.next_field.get_position(relative_to=camera.ui)
                invoke(setattr, self.next_field, 'active', True, delay=.01)

        if self.active and key in self.submit_on and self.on_submit:
            self.on_submit()
            self.active = False

        if self.active:
            self.text_field.input(key)
            return True # eat input when entering text to prevent unwanted actions while typing.


    def text_getter(self):
        return self.text_field.text

    def text_setter(self, value):
        self.text_field.text = ''
        self.text_field.cursor.position = (0,0)
        self.text_field.add_text(value, move_cursor=True)
        self.text_field.render()


    def text_color_getter(self):
        return self.text_field.text_entity.color
    def text_color_setter(self, value):
        self.text_field.text_entity.color = value

    def active_getter(self):
        return self.text_field.active
    def active_setter(self, value):
        self.text_field.active = value

    # def text_origin_setter(self, value):
    #     self.text_field.text_entity.text_origin = value


if __name__ == '__main__':
    app = Ursina()
    # window.fullscreen_size = (1366, 768)
    # background = Entity(model='quad', texture='pixelscape_combo', parent=camera.ui, scale=(camera.aspect_ratio,1), color=color.white)
    gradient = Entity(model='quad', texture='vertical_gradient', parent=camera.ui, scale=(camera.aspect_ratio,1), color=color.hsv(240,.6,.1,.75))

    username_field = InputField(y=-.12, limit_content_to='0123456789', default_value='11', active=True, scale_x=1)
    username_field.text = '0929468098'
    password_field = InputField(y=-.18, hide_content=True)
    username_field.next_field = password_field

    def submit():
        print('ursername:', username_field.text)
        print('password:',  password_field.text)

    Button(text='Login', scale=.1, color=color.cyan.tint(-.4), y=-.26, on_click=submit).fit_to_text()
    username_field.on_value_changed = submit
    app.run()



from ursina import *
app = Ursina()

draggable = Draggable(parent=scene, model='cube', plane_direction=Vec3.up, )
turret = Entity(model=Cone(), scale=Vec3(.5,1,.5), origin_y=-.5, color=color.azure,)
turret.model.colorize()
grid = Entity(model=Grid(8,8), scale=8, rotation_x=90)

def update():
    turret.lookAt(draggable.position, Vec3.up)

    draggable.y += (held_keys['e'] - held_keys['q']) * time.dt * 10

EditorCamera()
app.run()



import os
import platform
import subprocess
from copy import copy, deepcopy
from pathlib import Path
from ursina.mesh import Mesh
from ursina import application, color
from time import perf_counter
from ursina.string_utilities import print_info, print_warning
from ursina.vec3 import Vec3
import panda3d.core as p3d
import gltf
import builtins
from ursina.sequence import Func


imported_meshes = dict()
blender_scenes = dict()
# folders = (application.asset_folder, )

def load_model(name, folder=None, file_types=('.bam', '.ursinamesh', '.obj', '.glb', '.gltf', '.blend'), use_deepcopy=False, gltf_no_srgb=Func(getattr, application, 'gltf_no_srgb')):
    if callable(gltf_no_srgb):
        gltf_no_srgb = gltf_no_srgb()

    if not isinstance(name, str):
        raise TypeError(f"Argument name must be of type str, not {type(str)}")


    if '.' in name:
        full_name = name
        name = full_name.split('.')[0]
        file_types = ('.' + full_name.split('.',1)[1],)

    if name in imported_meshes:
        # print('load cached model', name)
        try:
            if not use_deepcopy:
                instance = copy(imported_meshes[name])
            else:
                instance = deepcopy(imported_meshes[name])

            instance.clearTexture()
            return instance

        except:
            pass

    if folder is not None:
        if not isinstance(folder, Path):
            raise TypeError(f'folder must be a Path, not a {type(folder)}')
        _folders = (folder,)

    else:
        _folders = (application.models_compressed_folder, application.asset_folder, application.internal_models_compressed_folder)


    for filetype in file_types:
        if use_deepcopy and filetype == '.bam':
            continue
        for folder in _folders:
            # warning: glob is case-insensitive on windows, so m.path will be all lowercase
            for file_path in folder.glob(f'**/{name}{filetype}'):
                if filetype == '.bam':
                    # print_info('loading bam')
                    return builtins.loader.loadModel(file_path)  # type: ignore

                if filetype == '.gltf' or filetype == '.glb':
                    gltf_settings = gltf.GltfSettings()
                    gltf_settings.no_srgb = gltf_no_srgb
                    model_root = gltf.load_model(str(file_path), gltf_settings=gltf_settings)
                    imported_meshes[name] = p3d.NodePath(model_root)
                    return p3d.NodePath(model_root)

                if filetype == '.ursinamesh':
                    try:
                        with open(file_path) as f:
                            m = eval(f.read())
                            m.path = file_path
                            m.name = name
                            m.vertices = [Vec3(*v) for v in m.vertices]
                            imported_meshes[name] = m
                            return m
                    except Exception as e:
                        raise Exception('invalid ursinamesh file:', file_path, e)


                if filetype == '.obj':
                    # print('found obj', file_path)
                    m = obj_to_ursinamesh(folder=folder, name=name, return_mesh=True)
                    m.path = file_path
                    m.name = name
                    imported_meshes[name] = m
                    if not use_deepcopy:
                        m.save(f'{name}.bam')

                    return m

                elif filetype == '.blend':
                    print_info('found blend file:', file_path)
                    if blend_to_obj(file_path):
                        m = obj_to_ursinamesh(folder=folder, name=name, return_mesh=True)
                        if not m:
                            raise ValueError('obj_to_ursinamesh failed to convert:', file_path)
                        m.path = file_path
                        m.name = name
                        imported_meshes[name] = m
                        if not use_deepcopy:
                            m.save(f'{name}.bam')

                        return m
                        # return load_model(name, folder, use_deepcopy=use_deepcopy)
                else:
                    try:
                        return builtins.loader.loadModel(file_path)  # type: ignore
                    except:
                        pass

    return None


# find blender installations
if application.development_mode:

    if platform.system() == 'Windows':
        # get blender path by getting default program for '.blend' file extension
        import shlex
        import winreg

        try:
            class_root = winreg.QueryValue(winreg.HKEY_CLASSES_ROOT, '.blend')
            with winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, fr'{class_root}\shell\open\command') as key:
                command = winreg.QueryValueEx(key, '')[0]
                default_blender = shlex.split(command)[0]
                default_blender = Path(default_blender)
                application.blender_paths['default'] = default_blender
                blender_foundation_directory = default_blender.parent.parent

                for blender_installation in blender_foundation_directory.glob('*'):
                    first_folder = tuple(blender_installation.glob('*'))[0] # version
                    version_name = first_folder.name[:3]
                    blender_path = list(blender_installation.glob('blender.exe'))[0]
                    if application.blender_path['default'] == blender_path:
                        continue
                    application.blender_paths[version_name] = blender_path
        except:
            pass

    elif platform.system() == 'Linux':
        # Use "which" command to find blender
        which_process = subprocess.run(('which', 'blender'), stdout=subprocess.PIPE)
        if which_process.returncode == 0:
            blender_exec = which_process.stdout.decode().strip()
            application.blender_paths['default'] = blender_exec


def load_blender_scene(name, folder:Path=Func(getattr, application, 'asset_folder'), reload=False, skip_hidden=True, models_only=False, uvs=True, vertex_colors=True, normals=True, triangulate=True, decimals=4):
    if callable(folder):
        folder = folder()

    scenes_folder = Path(application.asset_folder / 'scenes')
    if not scenes_folder.exists():
        scenes_folder.mkdir()

    out_file_path = scenes_folder / f'{name}.ursina_blender_scene'
    # print('loading:', out_file_path)
    if reload or not out_file_path.exists():
        print_info('reload:')
        blend_file = tuple(folder.glob(f'**/{name}.blend'))
        if not blend_file:
            raise ValueError('no blender file found at:', folder / name)

        blend_file = blend_file[0]
        print_info('loading blender scene:', blend_file, '-->', out_file_path)

        args = [
            get_blender(blend_file),
            blend_file,
            '--background',
            '--python',
            application.internal_scripts_folder / '_blender_scene_to_ursina.py',
            out_file_path,
            '--skip_hidden' if skip_hidden else '',
            '--models_only' if models_only else '',
            '--uvs' if uvs else '',
            '--normals' if normals else '',
            '--vertex_colors' if vertex_colors else '',
            '--triangulate' if triangulate else '',
            f'--decimals={decimals}',
        ]

        subprocess.run(args)


    with open(out_file_path) as f:
        file_content = f.read()
        loc = {}
        exec(file_content, globals(), loc)

        if models_only:
            blender_scenes[name] = loc['meshes']
            return loc['meshes']

        blender_scenes[name] = loc['scene_parent']
        return loc['scene_parent']



def get_blender(blend_file):    # try to get a matching blender version in case we have multiple blender version installed
    if not application.blender_paths:
        print_warning(f"Error: Trying to load .blend file, but no blender installation was found. blender_paths: {application.blender_paths}. If Blender is not installed, install it. If Blender is installed, but not found, make sure to install it to the default install location. If it's still not found, you can provide a custom path like this: application.blender_paths['default'] = Path('C:\Program Files\...')")
        return None

    if len(application.blender_paths) == 1:
        return application.blender_paths['default']

    with open(blend_file, 'rb') as f:
        try:
            blender_version_number = (f.read(12).decode("utf-8"))[-3:]   # get version from start of .blend file e.g. 'BLENDER-v280'
            blender_version_number = blender_version_number[0] + '.' + blender_version_number[1:2]
            print_info('blender_version:', blender_version_number)
            if blender_version_number in application.blender_paths:
                return application.blender_paths[blender_version_number]

            print_info('using default blender version')
            return application.blender_paths['default']
        except:
            print_info('using default blender version')
            return application.blender_paths['default']


def blend_to_obj(blend_file:Path, out_folder=Func(getattr, application, 'models_compressed_folder'), export_mtl=True):
    if callable(out_folder):
        out_folder = out_folder()

    if not out_folder.exists():
        out_folder.mkdir()

    export_script_path = application.internal_scripts_folder / '_blend_export.py'
    exported = []
    export_mtl_arg = '--export_mtl' if export_mtl else ''

    blender = get_blender(blend_file)
    if blender is None:
        raise Exception('trying to use blend_to_obj, but no blender installation was found. Please install blender to the default install location or provide a custom path in application.blender_paths.')

    name = f'{blend_file.stem}.obj'
    out_file_path = out_folder / name
    print_info('converting .blend file to .obj:', blend_file, '-->', out_file_path, 'using:', blender)

    if platform.system() == 'Windows':
        subprocess.call(f'''"{blender}" "{blend_file}" --background --python "{export_script_path}" {export_mtl_arg} "{out_file_path}"''')
    else:
        subprocess.run((blender, blend_file, '--background', '--python', export_script_path, out_file_path))

    exported.append(blend_file)

    return exported


def obj_to_ursinamesh(folder=Func(getattr, application, 'models_compressed_folder'), out_folder=Func(getattr, application, 'models_compressed_folder'), name='*', return_mesh=True, save_to_file=False, delete_obj=False):
    if callable(folder):
        folder = folder()
    if callable(out_folder):
        out_folder = out_folder()

    if name.endswith('.obj'):
        name = name[:-4]

    for file_path in folder.glob(f'**/{name}.obj'):
        # filepath = path / (os.path.splitext(f)[0] + '.obj')
        print('read obj at:', file_path)


        with file_path.open('r') as file:
            lines = file.readlines()

        verts = []
        tris = []

        uv_indices = []
        uvs = []
        norm_indices = []
        norms = []
        normals = [] # final normals made getting norms with norm_indices

        vertex_colors = []
        vertex_colors_packed = []
        current_color = None
        mtl_data = None
        mtl_dict = {}

        # parse the obj file to a Mesh
        for i, l in enumerate(lines):  # noqa: E741
            if l.startswith('v '):
                parts = [float(v) for v in l[2:].strip().split(' ')]
                vert = parts[:3]
                vert[0] = -vert[0]
                verts.append(tuple(vert))
                if len(parts) > 3:
                    # current_color = color.rgb(*parts[3:])
                    vertex_colors_packed.append(color.rgb(*parts[3:]))

            elif l.startswith('vn '):
                n = l[3:].strip().split(' ')
                norms.append(tuple(float(e) for e in n))

            elif l.startswith('vt '):
                uv = l[3:].strip()
                uv = uv.split(' ')
                uvs.append(tuple(float(e) for e in uv))

            elif l.startswith('f '):
                l = l[2:]  # noqa: E741
                l = l.split(' ')  # noqa: E741

                try:
                    tri = tuple(int(t.split('/')[0])-1 for t in l if t != '\n')
                except:
                    print_warning('error in obj file line:', i, ':', l)
                    return
                if len(tri) == 3:
                    tris.extend(tri)
                    if current_color:
                        vertex_colors.extend([current_color for i in range(3)])
                    elif vertex_colors_packed:
                        vertex_colors.extend([vertex_colors_packed[idx] if idx < len(vertex_colors_packed) else color.white for idx in tri])

                elif len(tri) == 4:
                    tris.extend((tri[0], tri[1], tri[2], tri[2], tri[3], tri[0]))
                    if current_color:
                        vertex_colors.extend([current_color for i in range(6)])
                    elif vertex_colors_packed:
                        vertex_colors.extend([vertex_colors_packed[idx] for idx in (tri[0], tri[1], tri[2], tri[2], tri[3], tri[0])])

                else: # ngon
                    for i in range(1, len(tri)-1):
                        tris.extend((tri[i], tri[i+1], tri[0]))
                    if current_color:
                        vertex_colors.extend([current_color for i in range(len(tri))])
                    elif vertex_colors_packed:
                        for i in range(1, len(tri)-1):
                            vertex_colors.extend([vertex_colors_packed[idx] for idx in (tri[i], tri[i+1], tri[0])])

                try:
                    uv = tuple(int(t.split('/')[1])-1 for t in l)
                    if len(uv) == 3:
                        uv_indices.extend(uv)
                    elif len(uv) == 4:
                        uv_indices.extend((uv[0], uv[1], uv[2], uv[2], uv[3], uv[0]))
                    else: # ngon
                        for i in range(1, len(uv)-1):
                            uv_indices.extend((uv[i], uv[i+1], uv[0]))
                except: # if no uvs
                    pass

                try:
                    n = tuple(int(t.split('/')[2])-1 for t in l)
                    if len(n) == 3:
                        norm_indices.extend(n)
                    elif len(uv) == 4:
                        norm_indices.extend((n[0], n[1], n[2], n[2], n[3], n[0]))
                    else: # ngon
                        for i in range(1, len(n)-1):
                            norm_indices.extend((n[i], n[i+1], n[0]))
                except: # if no normals
                    pass

            elif l.startswith('mtllib '):    # load mtl file
                mtl_file_name = file_path.with_suffix('.mtl')
                if mtl_file_name.exists():
                    with open(mtl_file_name, mode='r', encoding='utf-8') as mtl_file:
                        mtl_data = mtl_file.readlines()

                        for i in range(len(mtl_data)-1):
                            if mtl_data[i].startswith('newmtl '):
                                material_name = mtl_data[i].strip()[7:] # remove 'newmtl '
                                for j in range(i+1, min(i+8, len(mtl_data))):
                                    if mtl_data[j].startswith('newmtl'):
                                        break
                                    if mtl_data[j].startswith('Kd '):
                                        material_color = [float(e) for e in mtl_data[j].strip()[3:].split(' ')]
                                        mtl_dict[material_name] = *material_color, 1


            elif l.startswith('usemtl ') and mtl_data: # apply material color
                material_name = l[7:].strip()    # remove 'usemtl '
                if material_name in mtl_dict:
                    current_color = mtl_dict[material_name]


        if norms: # make sure we have normals and not just normal indices (weird edge case).
            normals = [(-norms[nid][0], norms[nid][1], norms[nid][2]) for nid in norm_indices]

        if return_mesh:
            return Mesh(
                vertices=[verts[t] for t in tris],
                normals=normals,
                uvs=[uvs[uid] for uid in uv_indices],
                colors=vertex_colors
                )

        mesh = Mesh(vertices=tuple(verts[t] for t in tris), colors=tuple(col for col in vertex_colors), uvs=tuple(uvs[uid] for uid in uv_indices), normals=normals)

        if not save_to_file:
            return mesh

        out_path = (out_folder / file_path.stem).with_suffix('.ursinamesh')
        # with open(out_path, 'w') as file:
        #     file.write(meshstring)
        mesh.save(folder=out_folder, name=f'{file_path.stem}.ursinamesh')

        if delete_obj:
            os.remove((out_folder / file_path.stem).with_suffix('.obj'))

        print_info('saved ursinamesh to:', out_path)

# faster, but does not apply modifiers
def blend_to_obj_fast(model_name=None, write_to_disk=False):
    print_info('find models')
    from tinyblend import BlenderFile  # type: ignore
    application.models_compressed_folder.mkdir(parents=True, exist_ok=True)

    files = os.listdir(application.models_folder)

    for f in files:
        if f.endswith('.blend'):
            # print('f:', application.models_compressed_folder + '/' + f)
            print('compress______', f)
            blend = BlenderFile(application.models_folder + '/' + f)
            number_of_objects = len(blend.list('Object'))

            for o in blend.list('Object'):
                if not o.data.mvert:
                    continue
                # print(o.id.name.decode("utf-8", "strict"))
                object_name = o.id.name.decode( "utf-8").replace(".", "_")[2:]
                object_name = object_name.split('\0', 1)[0]
                print('name:', object_name)

                verts = [v.co for v in o.data.mvert]
                verts = tuple(verts)

                file_content = 'Mesh(' + str(verts)

                file_name = ''.join([f.split('.')[0], '.ursinamesh'])
                if number_of_objects > 1:
                    file_name = ''.join([f.split('.')[0], '_', object_name, '.ursinamesh'])
                file_path = os.path.join(application.models_compressed_folder, file_name)
                print(file_path)

                tris = tuple(triindex.v for triindex in o.data.mloop)
                flippedtris = []
                for i in range(0, len(tris)-3, 3):
                    flippedtris.append(tris[i+2])
                    flippedtris.append(tris[i+1])
                    flippedtris.append(tris[i+0])

                file_content += ', triangles=' + str(flippedtris)

                if o.data.mloopuv:
                    uvs = tuple(v.uv for v in o.data.mloopuv)
                    file_content += ', uvs=' + str(uvs)

                file_content += ''', mode='triangle')'''

                if write_to_disk:
                    with open(file_path, 'w') as file:
                        file.write(file_content)

                return file_content

def ursina_mesh_to_obj(mesh, name='', out_path=Func(getattr, application, 'models_compressed_folder'), max_decimals=5, flip_faces=True):
    if callable(out_path):
        out_path = out_path()

    from ursina.string_utilities import camel_to_snake

    obj = ''
    obj += f'mtllib {name}.mtl\n'
    obj += f'usemtl {name}\n'

    if not name:
        name = camel_to_snake(mesh.__class__.__name__)
    obj += 'o ' + name + '\n'
    verts = mesh.vertices

    for v in verts:
        v = [round(e, max_decimals) for e in v]
        obj += f'v {v[0]} {v[1]} {v[2]}\n'

    if mesh.uvs:
        for uv in mesh.uvs:
            uv = [round(e, max_decimals) for e in uv]
            obj += f'vt {uv[0]} {uv[1]}\n'

    obj += 's off\n'

    if mesh.triangles:
        tris = mesh.triangles

        if isinstance(tris[0], tuple): # convert from tuples to flat
            new_tris = []
            for t in tris:
                if len(t) == 3:
                    if not flip_faces:
                        new_tris.extend([t[0], t[1], t[2]])
                    else:
                        new_tris.extend([t[2], t[1], t[0]])
                elif len(t) == 4: # turn quad into tris
                    if not flip_faces:
                        new_tris.extend([t[0], t[1], t[2], t[2], t[3], t[0]])
                    else:
                        new_tris.extend([t[2], t[1], t[0], t[0], t[3], t[2]])

            tris = new_tris


    if mesh.mode == 'ngon':
        tris = []
        for i in range(1, len(mesh.vertices)-1):
            tris.extend((i, i+1, 0))


    # tris must be a list of indices
    for i, t in enumerate(tris):
        if i % 3 == 0:
            obj += '\nf '
        obj += str(t+1)
        if mesh.uvs:
            obj += '/'+str(t+1)
        obj += ' '

    obj += '\n'
    # print(obj)
    with open(out_path / (name + '.obj'), 'w') as f:
        f.write(obj)
        print_info('saved obj:', out_path / (name + '.obj'))



def compress_internal():
    for blend_file in application.internal_models_folder.glob('*.blend'):
        blend_to_obj(blend_file, export_mtl=False)
        obj_to_ursinamesh(application.internal_models_compressed_folder, application.internal_models_compressed_folder, return_mesh=False, save_to_file=True, delete_obj=True)


if __name__ == '__main__':
    compress_internal()
    from ursina import *
    from ursina import Ursina, Entity, EditorCamera, Sky
    app = Ursina()
    # print('imported_meshes:\n', imported_meshes)
    # Entity(model='quad').model.save('quad.bam')
    m = obj_to_ursinamesh(folder=application.asset_folder.parent / 'samples', name='procedural_rock_0', save_to_file=False, delete_obj=False)
    print(m.serialize())
    # Entity(model=m)
    # EditorCamera()


    # application.asset_folder = application.asset_folder.parent / 'samples'
    # application.asset_folder = Path(r'C:\\Users\\Petter\\Downloads\\')
    # Entity(model='c1a0')
    # from ursina.shaders import lit_with_shadows_shader
    # Entity.default_shader = lit_with_shadows_shader
    # Entity(model='race')
    # Entity(model='ambulance', x=1.5)

    # application.asset_folder = Path(r'''C:\Users\Petter\Downloads''')
    # t = perf_counter()
    # Entity(model='untitled')
    # print('-------', perf_counter() - t)
    # m = load_model('cube', use_deepcopy=True)
    # ground = Entity(model='plane', scale=10, texture='brick', texture_scale=Vec2(4))
    # DirectionalLight()

    # blender_scene = load_blender_scene(path=application.asset_folder, name='desert', reload=True)
    # blender_scene = load_blender_scene(path=application.asset_folder, name='blender_level_editor_test_scene_2')
    # print('-------', time.time() - t)

    # print('--------', blender_scene.children)
    # for e in blender_scene.children:
    #     # e.color = color.random_color()
    #     e.shader = rim_shader
    #     e.texture='matcap_4'
    #
    #
    # blender_scene.Plane_002.collider = 'mesh'
    # from ursina.prefabs.first_person_controller import FirstPersonController
    # player = FirstPersonController()

    # def input(key):
    #     if key == '+':
    #         for e in blender_scene.children:
    #             e.texture_scale = Vec2(e.texture_scale[0], e.texture_scale[1]+.1)
    #     if key == '-':
    #         for e in blender_scene.children:
    #             e.texture_scale = Vec2(e.texture_scale[0], e.texture_scale[1]-.1)
    #         print(blender_scene.children[0].texture_scale)
    #
    EditorCamera()
    Sky(texture='sky_sunset')
    # def update():
    #     blender_scene.Cube.x += (held_keys['d'] - held_keys['a']) * time.dt * 10


    app.run()
    # e = Entity(model=Cylinder(16))
    # ursina_mesh_to_obj(e.model, name='quad_export_test')



from ursina import *


class Scrollable:

    def __init__(self, **kwargs):
        super().__init__()
        self.max = inf
        self.min = -inf
        self.scroll_speed = .05
        self.scroll_smoothing = 16
        self.axis = 'y'
        self.target_value = None

        for key, value in kwargs.items():
            setattr(self, key, value)



    def update(self):
        # lerp position
        if self.target_value:
            setattr(self.entity, self.axis, lerp(getattr(self.entity, self.axis), self.target_value, time.dt * self.scroll_smoothing))



    def input(self, key):
        if not mouse.hovered_entity:
            return

        if not self.target_value:
            self.target_value = getattr(self.entity, self.axis)

        if self.entity.hovered or mouse.hovered_entity.has_ancestor(self.entity):
            # print(key)
            if key == 'scroll up':
                self.target_value -= self.scroll_speed
            if key == 'scroll down':
                self.target_value += self.scroll_speed


            self.target_value = max(min(self.target_value, self.max), self.min)


if __name__ == '__main__':
    '''
    This will make target entity move up or down when you hover the entity/its children
    while scrolling the scroll wheel.
    '''

    app = Ursina()
    p = Button(model='quad', scale=(.4, .8), collider='box')
    for i in range(8):
        Button(parent=p , scale_y=.05, text=f'giopwjoigjwr{i}', origin_y=.5, y=.5-(i*.05))

    p.add_script(Scrollable())
    app.run()



from ursina import *


class Animator:
    def __init__(self, animations=None, start_state='', pause_disabled=True):

        self.animations = animations    # dict
        self.pause_disabled = pause_disabled

        if not start_state and self.animations:
            start_state = list(self.animations)[0]

        self.start_state = start_state
        self._state = None
        self.state = start_state



    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):

        if not value in self.animations:
            print_warning(self, 'has no animation:', value)

        elif not self._state == value:
            # only show set state and disable the rest
            for name, entity in self.animations.items():
                if entity:
                    entity.enabled = value == name
                    if self.pause_disabled and hasattr(entity, 'pause') and not entity.enabled:
                        [anim.pause() for anim in entity.animations]

            entity = self.animations[value]
            if entity:
                if hasattr(entity, 'start') and callable(entity.start):
                    entity.start()

                if hasattr(entity, 'animations'):
                    [anim.start() for anim in entity.animations]

        self._state = value




if __name__ == '__main__':
    app = Ursina()
    # texture_importer.textureless=True
    anim = Animation('ursina_wink', loop=True, autoplay=False)
    a = Animator(
        animations = {
            'lol' : Entity(model='cube', color=color.red),
            'yo' : Entity(model='cube', color=color.green, x=1),
            'help' : anim,
        }
    )
    a.state = 'yo'

    Text('press <red>1<default>, <green>2<default> or <violet>3<default> to toggle different animator states', origin=(0,-.5), y=-.4)

    def input(key):
        if key == '1':
            a.state = 'lol'
        if key == '2':
            a.state = 'yo'
        if key == '3':
            a.state = 'help'
            print(anim.enabled)

    app.run()



from ursina import Quat, Shader, Vec2, Vec3, Vec4

instancing_shader = Shader(name='instancing_shader', language=Shader.GLSL, vertex='''#version 140

uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
in vec4 p3d_Color;
out vec2 texcoords;
out vec4 vertex_color;
uniform vec2 texture_scale;
uniform vec2 texture_offset;

uniform vec3 position_offsets[256];
uniform vec4 rotation_offsets[256];
uniform vec3 scale_multipliers[256];

void main() {
    vec3 v = p3d_Vertex.xyz * scale_multipliers[gl_InstanceID];
    vec4 q = rotation_offsets[gl_InstanceID];
    v = v + 2.0 * cross(q.xyz, cross(q.xyz, v) + q.w * v);

    gl_Position = p3d_ModelViewProjectionMatrix * (vec4(v + position_offsets[gl_InstanceID], 1.));
    texcoords = (p3d_MultiTexCoord0 * texture_scale) + texture_offset;
    vertex_color = p3d_Color;
}
''',

fragment='''#version 140

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 texcoords;
in vec4 vertex_color;
out vec4 fragColor;

void main() {
    vec4 color = texture(p3d_Texture0, texcoords) * p3d_ColorScale * vertex_color;
    fragColor = color.rgba;
}
''',

default_input={
    'texture_scale': Vec2(1, 1),
    'texture_offset': Vec2(0.0, 0.0),
    'position_offsets': [Vec3(i, 0, 0) for i in range(256)],
    'rotation_offsets': [Vec4(0) for i in range(256)],
    'scale_multipliers': [Vec3(1) for i in range(256)],
}
)

if __name__ == '__main__':
    from ursina import Ursina, window, Entity, EditorCamera, Vec3, color, application, time, Cone
    import random
    window.vsync = False
    app = Ursina()

    instances = []
    Entity(model='plane', texture='grass', scale=128)
    application.asset_folder = application.asset_folder.parent.parent
    p = Entity(model=Cone(5), y=1, texture='brick')
    p.model.uvs = [(v[0],v[1]) for v in p.model.vertices]
    p.model.generate()
    p.shader = instancing_shader
    p.setInstanceCount(256)

    for z in range(16):
        for x in range(16):
            e = Entity(add_to_scene_entities=False, position=Vec3(x, 0, z), color=color.lime, rotation_y=random.random()*360)
            instances.append(e)
            print(e.quaternion, Quat())

    p.set_shader_input('position_offsets', [e.position*4 for e in instances])
    p.set_shader_input('rotation_offsets', [e.quaternion for e in instances])
    p.set_shader_input('scale_multipliers',[e.scale*random.uniform(.9,2) for e in instances])
    p.set_shader_input('colors', p.model.colors)

    print(len(p.model.vertices) * len(instances))
    EditorCamera()

    app.run()



from ursina import Entity, Vec2, invoke, scene, color, Vec3, print_warning
from ursina.prefabs.sky import Sky
from panda3d.core import DirectionalLight as PandaDirectionalLight
from panda3d.core import PointLight as PandaPointLight
from panda3d.core import AmbientLight as PandaAmbientLight
from panda3d.core import Spotlight as PandaSpotLight


class Light(Entity):
    default_values = {'rotation_x':90, }
    def __init__(self, **kwargs):
        super().__init__(**(__class__.default_values | kwargs))

    @property
    def color(self):
        return getattr(self, '_color', color.white)

    @color.setter
    def color(self, value):
        self._color = value
        self._light.setColor(value)

from ursina.scripts.property_generator import generate_properties_for_class


@generate_properties_for_class()
class DirectionalLight(Light):
    def __init__(self, shadow_map_resolution=Vec2(1024,1024), shadows=True, **kwargs):
        super().__init__(**kwargs)
        self._light = PandaDirectionalLight('directional_light')
        node_path = self.attachNewNode(self._light)
        node_path.node().setCameraMask(0b0001)
        render.setLight(node_path)

        self.shadow_map_resolution = shadow_map_resolution
        self._bounds_entity = scene
        self.shadows = shadows

    def shadows_setter(self, value):
        self._shadows = value
        if value:
            self._light.set_shadow_caster(True, int(self.shadow_map_resolution[0]), int(self.shadow_map_resolution[1]))
            self.update_bounds()
        else:
            self._light.set_shadow_caster(False)



    def update_bounds(self, entity=scene):  # update the shadow area to fit the bounds of target entity, defaulted to scene.
        # don't include skydome when calculating shadow bounds
        [e.disable() for e in Sky.instances]

        bounds = entity.get_tight_bounds(self)
        if bounds is not None:
            bmin, bmax = bounds
            lens = self._light.get_lens()
            lens.set_near_far(bmin.z*2, bmax.z*2)
            lens.set_film_offset((bmin.xy + bmax.xy) * .5)
            lens.set_film_size(bmax.xy - bmin.xy)
        else:
            lens = self._light.get_lens()
            lens.set_near_far(1, 100)
            lens.set_film_offset((0, 0))
            lens.set_film_size((1, 1))

        [e.enable() for e in Sky.instances]

        self._bounds_entity = entity

    def look_at(self, target, axis=Vec3.forward):
        super().look_at(target, axis=axis)
        self.update_bounds(self._bounds_entity)



class PointLight(Light):
    def __init__(self, **kwargs):
        super().__init__()
        self._light = PandaPointLight('point_light')
        render.setLight(self.attachNewNode(self._light))

        for key, value in kwargs.items():
            setattr(self, key ,value)



class AmbientLight(Light):
    def __init__(self, **kwargs):
        super().__init__()
        self._light = PandaAmbientLight('ambient_light')
        render.setLight(self.attachNewNode(self._light))

        for key, value in kwargs.items():
            setattr(self, key ,value)



class SpotLight(Light):
    def __init__(self, **kwargs):
        super().__init__()
        self._light = PandaSpotLight('spot_light')
        render.setLight(self.attachNewNode(self._light))

        for key, value in kwargs.items():
            setattr(self, key ,value)




if __name__ == '__main__':
    from ursina import Ursina, EditorCamera, color, Vec3
    from ursina.shaders import lit_with_shadows_shader # you have to apply this shader to enties for them to receive shadows.

    app = Ursina()

    Entity.default_shader = lit_with_shadows_shader
    ground = Entity(model='plane', scale=10, texture='grass')
    lit_cube = Entity(model='cube', y=1, color=color.light_gray)

    light = DirectionalLight()
    light.look_at(Vec3(1,-1,1))

    dont_cast_shadow = Entity(model='cube', y=1, x=2, color=color.light_gray)
    dont_cast_shadow.hide(0b0001)

    unlit_entity = Entity(model='cube', y=1,x=-2, unlit=True, color=color.light_gray)

    bar = Entity(model='cube', position=(0,3,-2), scale=(10,.2,.2), color=color.light_gray)
    # dont_cast_shadow.hide(0b0001)

    # How to render shows in a limited area.
    # to make it easier to see, make a box to define where we will have shadows. we can make this invisible after.
    shadow_bounds_box = Entity(model='wireframe_cube', scale=5, visible=0)
    light.update_bounds(shadow_bounds_box)

    EditorCamera(rotation=(30,30,0))
    Sky()
    # from ursina import window
    # window.borderless = True
    # window.size = Vec2(1920,1080) * .5
    # window.editor_ui.enabled = False
    app.run()



from ursina import camera, color, load_texture
from ursina.shader import Shader
from ursina.ursinastuff import Func
from ursina.vec2 import Vec2

fresnel_shader = Shader(name='fresnel_shader', language=Shader.GLSL,
vertex = '''
#version 140
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelMatrix;

in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 texcoord;
uniform vec2 texture_scale;
uniform vec2 texture_offset;

in vec3 p3d_Normal;
out vec3 world_normal;
out vec3 world_position;

void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    texcoord = (p3d_MultiTexCoord0 * texture_scale) + texture_offset;
    world_normal = normalize(mat3(p3d_ModelMatrix) * p3d_Normal);
    world_position = mat3(p3d_ModelMatrix) * p3d_Vertex.xyz;
}
''',

fragment='''
#version 140

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 texcoord;
in vec3 world_normal;
in vec3 world_position;
out vec4 fragColor;

uniform vec4 fresnel_color;
uniform sampler2D fresnel_texture;
uniform vec3 camera_world_position;
uniform float bias;
uniform float scale;
uniform float power;

vec3 do_fresnel(vec4 color) {
    vec3 offset_camera_world_position = camera_world_position + vec3(0.,0.,-1.);
    vec3 I = normalize(world_position - offset_camera_world_position.xyz);
    float fresnel = bias + scale * pow(1.0 + dot(I, world_normal), power);

    fresnel *= texture(fresnel_texture, texcoord).r;
    return mix(color.rgb, fresnel_color.rgb, fresnel*fresnel_color.a);
}

void main() {
    vec4 color = texture(p3d_Texture0, texcoord) * p3d_ColorScale;

    fragColor.rgb = do_fresnel(color);
    fragColor.a = color.a;
}


''', geometry='',
default_input = {
    'texture_scale' : Vec2(1,1),
    'texture_offset' : Vec2(0.0, 0.0),

    'fresnel_color' : color.light_gray,
    'fresnel_texture' : Func(load_texture, 'white_cube'),
    'bias' : .05,
    'scale' : 1,
    'power' : 2.0,
    'camera_world_position': Func(getattr, camera, 'world_position'),
},
continuous_input = {
    'camera_world_position': Func(getattr, camera, 'world_position'),
    }
)



if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    # window.color=color.black

    b = Entity(model='sphere', color=color.black, shader=fresnel_shader)
    b = Entity(model=Quad(), color=color.dark_gray, shader=fresnel_shader, x=.25, parent=camera.ui, scale=.2, ignore=True)
    b.shader_input = {
        'bias': .01,
        'scale': 1.5,
        'power': 1.5,
        'fresnel_color': color.hex('#123123')
    }
    b.animate_rotation_y(15, duration=1, curve=curve.linear_boomerang, loop=True)
    ground = Entity(model='plane', color=color.gray, shader=fresnel_shader, y=-1, scale=64, texture='grass', texture_scale=Vec2(32,32))
    ground.set_shader_input('fresnel_color', color.gray)
    ground.set_shader_input('fresnel_texture', load_texture('white_cube'))
    EditorCamera()

    app.run()



def singleton(cls):
    def get_instance(*args, **kwargs):
        if not hasattr(cls, '_singleton_instance') or not cls._singleton_instance:
            cls._singleton_instance = cls(*args, **kwargs)
        return cls._singleton_instance

    return get_instance


if __name__ == '__main__':
    from ursina.ursinastuff import _test

    class MyBaseClass:
        def __init__(self, name):
            self.name = name

    @singleton
    class DecoratedClass(MyBaseClass):
        def __init__(self, name='decorated_class'):
            super().__init__(name)

    app = DecoratedClass()
    app_2 = DecoratedClass()
    _test(app == app_2)

    # # this won't work
    # class Game(DecoratedClass):
    #     pass



from ursina import *
from ursina.networking import *

app = Ursina(borderless=False)

start_text = "Press H to host or C to connect."
status_text = Text(text=start_text, origin=(0, 0))
messages = []

peer = RPCPeer()

@rpc(peer)
def message(connection, time_received, msg: str):
    message = Text(text=f"Received: {msg}", origin=(0, 0), y=-0.05-len(messages)*0.05)
    messages.append(message)
    s = Sequence(1, Func(message.fade_out, duration=0.5), 0.5, Func(destroy, message), Func(messages.pop, 0))
    s.start()

def update():
    peer.update()

    if not peer.is_running():
        status_text.text = start_text
        return

    if peer.is_hosting():
        status_text.text = "Hosting on localhost, port 8080.\nSpace to send message."
    else:
        status_text.text = "Connected to host with address localhost, port 8080.\nSpace to send message."

def input(key):
    if key == "space":
        if peer.is_running():
            peer.message(peer.get_connections()[0], "Hello, World!")
    elif key == "h":
        peer.start("localhost", 8080, is_host=True)
    elif key == "c":
        peer.start("localhost", 8080, is_host=False)

app.run()
 



def normalize_v3(arr):
    ''' Normalize a numpy array of 3 component vectors shape=(n,3) '''
    import numpy

    lens = numpy.sqrt( arr[:,0]**2 + arr[:,1]**2 + arr[:,2]**2 )
    arr[:,0] /= lens
    arr[:,1] /= lens
    arr[:,2] /= lens
    return arr


def generate_normals(vertices, triangles=None, smooth=True):
    import numpy

    if not vertices:
        raise ValueError("can't generate normals for a mesh with 0 vertices")

    if not triangles:
        new_tris = [(i, i+1, i+2) for i in range(0, len(vertices), 3)]

    elif not isinstance(triangles[0], int):
        raise TypeError(f'triangles must be ints, not {type(triangles[0])} ({triangles[0]})')

    else:
        new_tris = [(triangles[i], triangles[i+1], triangles[i+2]) for i in range(0, len(triangles), 3)]


    vertices = numpy.array(vertices)
    triangles = numpy.array(new_tris)

    normals = numpy.zeros(vertices.shape, dtype=vertices.dtype)
    #Create an indexed view into the vertex array using the array of three indices for triangles
    tris = vertices[triangles]
    #Calculate the normal for all the triangles, by taking the cross product of the vectors v1-v0, and v2-v0 in each triangle
    n = numpy.cross(tris[::,1] - tris[::,0] ,tris[::,2] - tris[::,0])
    # n is now an array of normals per triangle. The length of each normal is dependent the vertices,
    # we need to normalize these, so that our next step weights each normal equally.
    normalize_v3(n)

    # inverse it, dunno why
    n = [-e for e in n]

    # now we have a normalized array of normals, one per triangle, i.e., per triangle normals.
    # But instead of one per triangle (i.e., flat shading), we add to each vertex in that triangle,
    # the triangles' normal. Multiple triangles would then contribute to every vertex, so we need to normalize again afterwards.
    # The cool part, we can actually add the normals through an indexed view of our (zeroed) per vertex normal array
    normals[triangles[:,0]] += n
    normals[triangles[:,1]] += n
    normals[triangles[:,2]] += n
    normalize_v3(normals)

    # smooth
    if smooth:
        vertices = vertices.tolist()
        visited = set()

        for i, v in enumerate(vertices):
            if i in visited:
                continue

            overlapping_verts_indices = [j for j, w in enumerate(vertices) if w == v and j not in visited]
            visited.update(overlapping_verts_indices)

            if overlapping_verts_indices:
                average_normal = sum(normals[e] for e in overlapping_verts_indices) / 3
                for index in overlapping_verts_indices:
                    normals[index] = average_normal


    return normals

if __name__ == '__main__':
    vertices = (
        (-0.0, -0.5, 0.0), (0.1, -0.48, -0.073), (-0.038, -0.48, -0.11),
        (0.361804, -0.22, -0.26), (0.3, -0.32, -0.22), (0.40, -0.25, -0.14),
        (-0.0, -0.5, 0.0), (-0.038, -0.48, -0.11), (-0.03, -0.48, -0.11)
    )
    from time import perf_counter
    t = perf_counter()
    norms = generate_normals(vertices, smooth=True)
    print('------', perf_counter() - t)
    # print(norms)
    # from ursina import *
    # app = Ursina()
    # m = Mesh(vertices=vertices)
    # m.generate_normals()
    # e = Entity(model=m)
    # # print(e.normals)
    # if e.normals:
    #     verts = list()
    #     for i in range(len(e.vertices)):
    #         verts.append(e.vertices[i])
    #         verts.append(Vec3(e.vertices[i][0], e.vertices[i][1], e.vertices[i][2])
    #             + Vec3(e.normals[i][0], e.normals[i][1], e.normals[i][2])*2)
    #
    #     lines=Entity(model=Mesh(verts, mode='line'))
    # # e.shader = 'shader_normals'
    # EditorCamera()
    # app.run()



from ursina import *


class Tooltip(Text):

    def __init__(self, text='', wordwrap=40, background_color=color.black66, **kwargs):
        super().__init__(text=text, ignore=False, parent=camera.ui, wordwrap=wordwrap, origin=(-.5,-.5), margin=(2,2),
            background_color=color.inverse(color.text_color), enabled=False)

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.create_background()
        self.background.color = background_color


    def update(self):
        self.position = mouse.position
        self.position = (
            mouse.x + (self.margin[0] * self.size/2) + .01,
            mouse.y + (self.margin[1] * self.size/2) + .01
            )
        self.x = min(self.x, (.5 * window.aspect_ratio) - self.width - self.size - .005)
        self.y = min(self.y, .5 - (self.height + self.size + .005))
        self.z = -99

if __name__ == '__main__':
    app = Ursina()

    tooltip_test = Tooltip(
    '<scale:1.5><pink>' + 'Rainstorm' + '<scale:1> \n \n' +
'''Summon a <blue>rain
storm <default>to deal 5 <blue>water
damage <default>to <red>everyone, <default>including <orange>yourself. <default>
Lasts for 4 rounds.'''.replace('\n', ' '),
        background_color=color.violet,
        font=Text.default_monospace_font,
        wordwrap=50,
)

    tooltip_test.enabled = True
    app.run()



from ursina import *

app = Ursina()

window.color = color.black
camera.orthographic = True
camera.fov = 1

left_paddle = Entity(scale=(1/32,6/32), x=-.75, model='quad', origin_x=.5, collider='box')
right_paddle = duplicate(left_paddle, x=left_paddle.x*-1, rotation_z=left_paddle.rotation_z+180)

floor = Entity(model='quad', y=-.5, origin_y=.5, collider='box', scale=(2,10), visible=False)
ceiling = duplicate(floor, y=.5, rotation_z=180, visible=False)
left_wall = duplicate(floor, x=-.5*window.aspect_ratio, rotation_z=90, visible=True)
right_wall = duplicate(floor, x=.5*window.aspect_ratio, rotation_z=-90, visible=True)

# Score variables
left_score = 0
right_score = 0
max_score = 5
game_paused = False  # Added variable to track the game's paused state

collision_cooldown = .15
ball = Entity(model='circle', scale=.05, collider='box', speed=0, collision_cooldown=collision_cooldown)

# Score text
score_text = Text(text=f"{left_score} : {right_score}", position=(0, .45), scale=2, origin=(0, 0))

def update():
    global left_score, right_score, game_paused

    if game_paused:
        return  # Stop all updates if the game is paused

    ball.collision_cooldown -= time.dt
    ball.position += ball.right * time.dt * ball.speed

    left_paddle.y += (held_keys['w'] - held_keys['s']) * time.dt * 1
    right_paddle.y += (held_keys['up arrow'] - held_keys['down arrow']) * time.dt * 1

    if ball.collision_cooldown > 0:
        return

    hit_info = ball.intersects()
    if hit_info.hit:
        ball.collision_cooldown = collision_cooldown

        if hit_info.entity in (left_paddle, right_paddle):
            ball.rotation_z += 180 * (-1 if hit_info.entity == left_paddle else 1)
            ball.rotation_z -= (hit_info.entity.world_y - ball.y) * 20 * 32 * (-1 if hit_info.entity == left_paddle else 1)
            ball.speed *= 1.1

        elif hit_info.entity == right_wall:
            left_score += 1
            update_score()

        elif hit_info.entity == left_wall:
            right_score += 1
            update_score()

        # Particle effect on collision
        particle = Entity(model='quad', position=hit_info.world_point, scale=0, texture='circle', add_to_scene_entities=False)
        particle.animate_scale(.2, .5, curve=curve.out_expo)
        particle.animate_color(color.clear, duration=.5, curve=curve.out_expo)
        destroy(particle, delay=.5)

    # Ball bounces off top and bottom of the screen
    if ball.y > ceiling.y - ball.scale_y/2 or ball.y < floor.y + ball.scale_y/2:
        ball.rotation_z = -ball.rotation_z  # Reverse vertical direction

def update_score():
    global left_score, right_score, game_paused
    score_text.text = f"{left_score} : {right_score}"

    if left_score >= max_score or right_score >= max_score:
        winner_text = Text(f"{'Left' if left_score >= max_score else 'Right'} Player Wins!", y=0, scale=2, origin=(0, 0))
        ball.speed = 0
        game_paused = True  # Pause the game after a win
        invoke(destroy, winner_text, delay=3)
    else:
        reset()

def reset():
    ball.position = (0, 0, 0)
    ball.rotation = (0, 0, 0)
    ball.speed = 10
    for paddle in (left_paddle, right_paddle):
        paddle.collision = True
        paddle.y = 0

info_text = Text("press space to play", y=-.45)

def input(key):
    global game_paused

    if key == 'space' and not game_paused:  # Prevent restarting when the game is paused
        info_text.enabled = False
        reset()

app.run()



from ursina import *


def texture_to_height_values(heightmap:Texture, skip=1):
    from numpy import asarray, flip, swapaxes
    from PIL import Image

    heightmap = heightmap
    skip = skip    # should be power of two. only works with heightmap, not height_values.

    if not isinstance(heightmap, Texture):
        heightmap = load_texture(heightmap)
        if not heightmap:
            print('failed to load heightmap:', heightmap)
            return

    width, depth = heightmap.width//skip, heightmap.height//skip
    img = heightmap.to_PIL_Image().convert('L')
    if skip > 1:
        img = img.resize([width, depth], Image.LANCZOS)

    height_values = asarray(img)
    height_values = flip(height_values, axis=0)
    height_values = swapaxes(height_values, 0, 1)
    return height_values



class Terrain(Mesh):
    def __init__(self, heightmap='', height_values=None, gradient=None, skip=1, **kwargs):

        if heightmap:
            self.height_values = texture_to_height_values(heightmap, skip)

        elif height_values is not None:
            self.height_values = height_values


        self.width = len(self.height_values)
        self.depth = len(self.height_values[0])
        self.aspect_ratio = self.width / self.depth
        self.gradient = gradient
        super().__init__()
        self.generate()


    def generate(self):
        # copy this from Plane to avoid unnecessary init
        self.vertices = []
        self.triangles = []
        self.uvs = []
        self.normals = []

        _height_values = [[j/255 for j in i] for i in self.height_values]
        w, h = self.width, self.depth
        centering_offset = Vec2(-.5, -.5)

        # create the plane
        i = 0
        for z in range(h):
            for x in range(w):

                self.vertices.append(Vec3((x/(w-1))+(centering_offset.x), _height_values[x][z], (z/(h-1))+centering_offset.y))
                self.uvs.append((x/w, z/h))

                if x > 0 and z > 0:
                    self.triangles.append((i, i-1, i-w-1, i-w-0))

                # normals
                if x > 0 and z > 0 and x < w-1 and z < h-1:
                    rl =  _height_values[x+1][z] - _height_values[x-1][z]
                    fb =  _height_values[x][z+1] - _height_values[x][z-1]
                    self.normals.append(Vec3(rl, 1, fb).normalized())
                else:
                    self.normals.append(Vec3(0,1,0))

                i += 1

        if self.gradient:
            self.colors = []
            for z, column in enumerate(self.height_values):
                for x, row in enumerate(column):
                    self.vertices.append(Vec3(x/w, self.height_values[x][z], z/h) + Vec3(centering_offset.x, 0, centering_offset.y))
                    y = int(self.height_values[x][z]*1)
                    y = clamp(y, 0, 255)
                    self.colors.append(self.gradient[y])


        super().generate()



if __name__ == '__main__':
    app = Ursina()
    '''Terrain using an RGB texture as input'''
    terrain_from_heightmap_texture = Entity(model=Terrain('heightmap_1', skip=8), scale=(40,5,20), texture='heightmap_1')

    '''
    I'm just getting the height values from the previous terrain as an example, but you can provide your own.
    It should be a list of lists, where each value is between 0 and 255.
    '''
    hv = terrain_from_heightmap_texture.model.height_values.tolist()
    terrain_from_list = Entity(model=Terrain(height_values=hv), scale=(40,5,20), texture='heightmap_1', x=40)
    terrain_bounds = Entity(model='wireframe_cube', origin_y=-.5, scale=(40,5,20), color=color.lime)

    def input(key):
        if key == 'space':  # randomize the terrain
            terrain_from_list.model.height_values = [[random.uniform(0,255) for a in column] for column in terrain_from_list.model.height_values]
            terrain_from_list.model.generate()

    EditorCamera(rotation_x=90)
    camera.orthographic = True
    Sky()
    player = Entity(model='sphere', color=color.azure, scale=.2, origin_y=-.5)

    def update():
        direction = Vec3(held_keys['d'] - held_keys['a'], 0, held_keys['w'] - held_keys['s']).normalized()
        player.position += direction * time.dt * 8
        y = terraincast(player.world_position, terrain_from_list, terrain_from_list.model.height_values)
        if y is not None:
            player.y = y





    app.run()



from ursina import *
from math import floor


app = Ursina()

size = 32
plane = Entity(model='quad', color=color.azure, origin=(-.5,-.5), z=10, collider='box', scale=size) # create an invisible plane for the mouse to collide with
grid = [[Entity(model='quad', position=(x,y), texture='white_cube', enabled=False) for y in range(size)] for x in range(size)] # make 2d array of entities
player = Entity(model='quad', color=color.orange, position=(16,16,-.1))
cursor = Entity(model=Quad(mode='line'), color=color.lime)

def update():
    if mouse.hovered_entity == plane:
        # round the cursor position
        cursor.position = mouse.world_point
        cursor.x = round(cursor.x, 0)
        cursor.y = round(cursor.y, 0)

    # check the grid if the player can move there
    if held_keys['a'] and not grid[int(player.x)][int(player.y)].enabled:
        player.x -= 5 * time.dt
    if held_keys['d'] and not grid[int(player.x)+1][int(player.y)].enabled:
        player.x += 5 * time.dt



def input(key):
    if key == 'left mouse down':
        grid[int(cursor.x)][int(cursor.y)].enabled = True
    if key == 'right mouse down':
        grid[int(cursor.x)][int(cursor.y)].enabled = False


camera.orthographic = True
camera.fov = 10
camera.position = (16,18)

# enable all tiles lower than 16 to make a ground
for column in grid:
    for e in column:
        if e.y <= 15:
            e.enabled = True

app.run()



from ursina import color
from ursina.shader import Shader

colored_lights_shader = Shader(name='colored_lights_shader', language=Shader.GLSL,
vertex='''
#version 130
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelMatrix;
in vec4 p3d_Vertex;
in vec3 p3d_Normal;
in vec4 p3d_Color;
in vec2 p3d_MultiTexCoord0;

out vec2 uv;
out vec3 world_normal;
out vec4 vertex_color;


void main() {
  gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
  world_normal = normalize(mat3(p3d_ModelMatrix) * p3d_Normal);
  vertex_color = p3d_Color;
  uv = p3d_MultiTexCoord0;
}
''',

fragment='''
#version 130

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 uv;
in vec3 world_normal;
in vec4 vertex_color;

uniform vec4 top_color;
uniform vec4 bottom_color;
uniform vec4 left_color;
uniform vec4 right_color;
uniform vec4 front_color;
uniform vec4 back_color;

out vec4 fragColor;


void main() {
    vec4 norm = vec4(world_normal*0.5+0.5, 1);
    vec4 color = texture(p3d_Texture0, uv) * vertex_color * p3d_ColorScale;
    // color += mix(bottom_color, top_color, norm.y) / 3.;
    // color += mix(right_color, left_color, norm.x) / 3.;
    // color += mix(front_color, back_color, norm.z) / 3.;
    color *= mix(bottom_color, top_color, norm.y);
    color *= mix(left_color, right_color, norm.x) * 1.5;
    color *= mix(front_color, back_color, norm.z) * 2.;
    // vec3 target_col = vec3(
    //     mix(right_color, left_color, norm.x),
    //     mix(bottom_color, top_color, norm.y),
    //     mix(front_color, back_color, norm.z)
    // );

    // color.rgb -= vec3(.5,.5,.5);
    // color.rgb = mix(color, target_col, .5);
    color = mix(color, vec4(1), 0.1);
    color *= 1.25;
    fragColor = vec4(color.rgb, 1.);
    // fragColor = vec4(vcol.rgb, 1.);
}


''',
geometry='',
default_input = {
    'top_color': color.hsv(220, .12, .82),
    'bottom_color': color.hsv(285, .13, .47),
    'left_color': color.hsv(217, .3, .68),
    'right_color': color.hsv(0, .25, .93),
    'front_color': color.hsv(231, .08, .69),
    'back_color': color.hsv(240, .05, .76),
    }
)


if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina()
    window.color=color.black

    # e = Entity(model='sphere', shader=basic_lighting_shader)
    shader = colored_lights_shader

    Entity(model='cube', color=color.white, shader=colored_lights_shader)
    e = Entity(model='cube', x=1.2, shader=colored_lights_shader, color=color.white)
    e.set_shader_input('top_color', hsv(0,1,1))
    e.set_shader_input('bottom_color', hsv(0,0,0))
    e.set_shader_input('left_color', hsv(0,0,0))
    e.set_shader_input('right_color', hsv(0,0,0))
    e.set_shader_input('front_color', hsv(0,0,0))
    e.set_shader_input('back_color', hsv(0,0,0))

    GrayPlane(scale=10, y=-2, texture='shore')

    Sky(color=color.light_gray)
    EditorCamera(rotate_around_mouse_hit=False)


    EditorCamera()

    app.run()



from ursina.vec2 import Vec2
from ursina.vec3 import Vec3
from ursina.vec4 import Vec4

import panda3d.core as p3d

from direct.distributed.PyDatagram import PyDatagram
from direct.distributed.PyDatagramIterator import PyDatagramIterator

from enum import Enum, auto

from collections import deque
import queue

import uuid
import hashlib

import types
import typing
import inspect

import socket
import ssl
import select
import errno

import struct

import time

import atexit
import signal

import threading
import asyncio

# Used internally by Peer.
class PeerEvent(Enum):
    ERROR = auto()
    CONNECT = auto()
    DISCONNECT = auto()
    DATA = auto()


# This class represents and single connection.
# It can be hashed and compared so that it may be used as a dictionary key.
# This is useful for mapping from a connection to player data.
class Connection:
    def __init__(self, peer, socket, address, connection_timeout):
        self.peer = peer
        self.socket = socket
        self.address = address

        self.connection_timeout = connection_timeout

        self.connected = True
        self.timed_out = False

        self.state = "l"
        self.length_byte_count = 2
        self.expected_byte_count = self.length_byte_count
        self.bytes_received = bytearray()

        self.uid = str(uuid.uuid4())

        self.receiving_thread = None

    def __hash__(self):
        return hash(self.uid)

    def __eq__(self, other):
        return self.uid == other.uid

    def send(self, data):
        b = bytearray()
        b += struct.pack(">H", len(data))
        b += data
        try:
            self.socket.sendall(b)
        except:
            pass

    def disconnect(self):
        if self.connected:
            try:
                self.socket.shutdown(socket.SHUT_RDWR)
                self.socket.close()
            except:
                pass
            self.connected = False
            self.peer._remove_connection(self)

    def is_timed_out(self):
        return self.timed_out

    def is_connected(self):
        return self.connected

    def _receive(self):
        try:
            while True:
                data = None
                try:
                    if self.connection_timeout is not None:
                        self.socket.settimeout(self.connection_timeout)
                    data = self.socket.recv(self.expected_byte_count)
                except socket.timeout:
                    self.timed_out = True
                    raise
                except Exception as e:
                    raise
                if data is None:
                    break
                if len(data) == 0:
                    break
                self.bytes_received += data
                self.expected_byte_count -= len(data)

                if self.expected_byte_count == 0:
                    if self.state == "l":
                        l = struct.unpack(">H", self.bytes_received)[0]
                        self.state = "c"
                        self.expected_byte_count = l
                        self.bytes_received.clear()
                    elif self.state == "c":
                        d = bytes(self.bytes_received.copy())
                        self.peer.output_event_queue.put((PeerEvent.DATA, self, d, time.time()))
                        self.state = "l"
                        self.expected_byte_count = self.length_byte_count
                        self.bytes_received.clear()
        except:
            pass
        finally:
            self.disconnect()
            self.state = "l"
            self.bytes_received.clear()
            self.expected_byte_count = self.length_byte_count


# -- Description --
# The main driving class of the networking module.
# This is either a server or a client depending on if it's hosting or not.
# -- Callbacks --
# The arguments starting with "on" are callback functions for network events.
# Their arguments are generally (connection, data, time).
# The connect and disconnect functions don't have the data argument.
# See the networking samples or update function definition for more.
# -- TLS / secure networking --
# TLS support can be enabled with `use_tls=True`.
# The host requires the path to the certificate chain, and the private key.
# The client can make use of a given path to a certificate authority bundle for testing / development / self signed.
# -- Address family --
# The socket address family can be either "INET" (ipv4) or "INET6" (ipv6).
# -- Notes --
# Keep in mind that the networking in running on its own thread and you must therefore check if it's running (is_running).
class Peer:
    def __init__(self, on_connect=None, on_disconnect=None, on_data=None, on_raw_data=None,
                 connection_timeout=None,
                 use_tls=False,
                 path_to_certchain=None, path_to_private_key=None,
                 path_to_cabundle=None,
                 socket_address_family="INET"):
        self.on_connect = on_connect
        self.on_disconnect = on_disconnect
        self.on_data = on_data
        self.on_raw_data = on_raw_data
        self.connection_timeout = connection_timeout
        self.use_tls = use_tls
        self.path_to_certchain = path_to_certchain
        self.path_to_private_key = path_to_private_key
        self.path_to_cabundle = path_to_cabundle
        if socket_address_family == "INET":
            self.socket_address_family = socket.AF_INET
        elif socket_address_family == "INET6":
            self.socket_address_family = socket.AF_INET6
        else:
            raise Exception("Invalid/unsupported socket address family '{socket_address_family}'.")

        self.ssl_context = None

        self.connections = []

        self.output_event_queue = queue.Queue()

        self.socket = None
        self.listen_socket = None
        self.host_name = None
        self.tls_host_name = None
        self.port = None
        self.backlog = 100
        self.is_host = False

        self.running = False

        self.main_thread = None
        self.server_listen_thread = None
        self.running_lock = threading.Lock()
        self.connections_lock = threading.Lock()

        def on_application_exit():
            if self.running:
                self.stop()

        atexit.register(on_application_exit)

        prev_signal_handler = signal.getsignal(signal.SIGINT)

        def on_keyboard_interrupt(*args):
            self.stop()
            if prev_signal_handler is not None:
                if prev_signal_handler != signal.SIG_DFL:
                    prev_signal_handler(*args)

        signal.signal(signal.SIGINT, on_keyboard_interrupt)

    def is_using_tls(self):
        return self.use_tls

    def start(self, host_name, port, is_host=False, backlog=100, tls_host_name=None, socket_address_family=None):
        if self.running:
            self.stop()

        if socket_address_family is not None:
            if socket_address_family == "INET":
                self.socket_address_family = socket.AF_INET
            elif socket_address_family == "INET6":
                self.socket_address_family = socket.AF_INET6
            else:
                raise Exception(f"Invalid/unsupported socket address family '{socket_address_family}'.")

        if is_host:
            if self.use_tls:
                self.ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
                self.ssl_context.load_cert_chain(self.path_to_certchain, self.path_to_private_key)
        else:
            if self.use_tls:
                if self.path_to_cabundle is not None:
                    self.ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                    self.ssl_context.load_verify_locations(self.path_to_cabundle)
                else:
                    self.ssl_context = ssl.create_default_context()

        self.host_name = host_name
        self.tls_host_name = tls_host_name
        if self.tls_host_name is None and self.use_tls:
            self.tls_host_name = self.host_name
        self.port = port
        self.backlog = backlog
        self.is_host = is_host

        self.main_thread = threading.Thread(target=self._run, daemon=True)
        self.main_thread.start()

    def stop(self):
        if not self.running:
            return

        self.disconnect_all()

        with self.running_lock:
            self.running = False

        if self.is_host:
            try:
                self.listen_socket.shutdown(socket.SHUT_RDWR)
                self.listen_socket.close()
            except:
                raise

        self.main_thread.join()

        while not self.output_event_queue.empty():
            self.output_event_queue.get()

    def update(self, max_events=100):
        for i in range(max_events):
            if self.output_event_queue.empty():
                break
            next_event = self.output_event_queue.get()
            event_type = next_event[0]
            conn = next_event[1]
            d = next_event[2]
            t = next_event[3]
            if event_type == PeerEvent.CONNECT:
                if self.on_connect is not None:
                    self.on_connect(conn, t)
            elif event_type == PeerEvent.DISCONNECT:
                if self.on_disconnect is not None:
                    self.on_disconnect(conn, t)
            elif event_type == PeerEvent.DATA:
                if self.on_raw_data is not None:
                    d = self.on_raw_data(conn, d, t)
                if self.on_data is not None:
                    self.on_data(next_event[1], d, t)

    def send(self, connection, data):
        connection.send(data)

    def disconnect(self, connection):
        connection.disconnect()

    def disconnect_all(self):
        connections = self.get_connections()
        for connection in connections:
            connection.disconnect()

    def is_running(self):
        return self.running

    def is_hosting(self):
        return self.is_host

    def connection_count(self):
        return len(self.connections)

    def get_connections(self):
        connections_copy = None
        with self.connections_lock:
            connections_copy = self.connections.copy()
        return connections_copy

    def _add_connection(self, socket, address):
        connection = Connection(self, socket, address, self.connection_timeout)
        with self.connections_lock:
            self.connections.append(connection)
        self.output_event_queue.put((PeerEvent.CONNECT, connection, None, time.time()))
        connection.receiving_thread = threading.Thread(target=connection._receive, daemon=True)
        connection.receiving_thread.start()

    def _remove_connection(self, connection):
        with self.connections_lock:
            if connection in self.connections:
                self.connections.remove(connection)
                self.output_event_queue.put((PeerEvent.DISCONNECT, connection, None, time.time()))

    def _run_server(self):
        try:
            self.listen_socket = socket.socket(self.socket_address_family, socket.SOCK_STREAM, 0)
            self.listen_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.listen_socket.bind((self.host_name, self.port))
            self.listen_socket.listen(self.backlog)
            if self.use_tls:
                self.listen_socket = self.ssl_context.wrap_socket(self.listen_socket, server_side=True)
        except Exception as e:
            print(e)
            raise

        with self.running_lock:
            self.running = True

        while self.running:
            try:
                client_socket, address = self.listen_socket.accept()
                client_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            except:
                break
            self._add_connection(client_socket, address)

    def _run_client(self):
        client_socket = None
        if self.use_tls:
            try:
                self.socket = socket.socket(self.socket_address_family, socket.SOCK_STREAM, 0)
                client_socket = self.ssl_context.wrap_socket(self.socket, server_hostname=self.tls_host_name)
                client_socket.connect((self.host_name, self.port))
            except Exception as e:
                self.running = False
                return
        else:
            try:
                client_socket = socket.create_connection((self.host_name, self.port))
            except Exception as e:
                self.running = False
                return
        client_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self._add_connection(client_socket, (self.host_name, self.port))

        with self.running_lock:
            self.running = True

    def _run(self):
        if self.is_host:
            self._run_server()
        else:
            self._run_client()


# -- Description --
# Main serialization class used by the networking module.
class DatagramWriter:
    def __init__(self):
        self.datagram = PyDatagram()
        self.builtin_type_functions = {
            bool: self.write_bool,
            int: self.write_int64,
            float: self.write_float64,
            str: self.write_string32,
            Vec2: self.write_vec2,
            Vec3: self.write_vec3,
            Vec4: self.write_vec4,
            tuple: self.write_tuple,
            list: self.write_list,
            bytes: self.write_blob
        }
        self.extra_type_functions = dict()

    def register_type(self, the_type, write_func):
        self.extra_type_functions[the_type] = write_func

    def clear(self):
        self.datagram = PyDatagram()

    def get_datagram(self):
        return self.datagram

    def write(self, value):
        type_of_value = type(value)
        builtin_converter_func = self.builtin_type_functions.get(type_of_value)
        if builtin_converter_func is not None:
            builtin_converter_func(value)
        else:
            converter_func = self.extra_type_functions.get(type_of_value)
            if converter_func is not None:
                converter_func(self, value)
            else:
                raise Exception(f"Unsupported value type for DatagramWriter: {type(value).__name__}")

    def write_string(self, value):
        self.datagram.addString(value)

    def write_string32(self, value):
        self.datagram.addString32(value)

    def write_bool(self, value):
        self.datagram.addBool(value)

    def write_int8(self, value):
        self.datagram.addInt8(value)

    def write_int16(self, value):
        self.datagram.addBeInt16(value)

    def write_int32(self, value):
        self.datagram.addBeInt32(value)

    def write_int64(self, value):
        self.datagram.addBeInt64(value)

    def write_float32(self, value):
        self.datagram.addBeFloat32(value)

    def write_float64(self, value):
        self.datagram.addBeFloat64(value)

    def write_blob(self, value):
        self.datagram.addBlob(value)

    def write_blob32(self, value):
        self.datagram.addBlob32(value)

    def write_vec2(self, value):
        self.write_float64(value[0])
        self.write_float64(value[1])

    def write_vec3(self, value):
        self.write_float64(value[0])
        self.write_float64(value[1])
        self.write_float64(value[2])

    def write_vec4(self, value):
        self.write_float64(value[0])
        self.write_float64(value[1])
        self.write_float64(value[2])
        self.write_float64(value[3])

    def write_tuple(self, value):
        for v in value:
            self.write(v)

    def write_list(self, value):
        self.write_int16(len(value))
        for v in value:
            self.write(v)


# Used internally by networking module.
class ExceedsListLimitException(Exception):
    pass


# -- Description --
# Main deserialization class used by the networking module.
class DatagramReader:
    def __init__(self):
        self.datagram = None
        self.iter = None
        self.builtin_read_functions = {
            bool: self.read_bool,
            int: self.read_int64,
            float: self.read_float64,
            str: self.read_string32,
            Vec2: self.read_vec2,
            Vec3: self.read_vec3,
            Vec4: self.read_vec4,
            bytes: self.read_blob

        }
        self.extra_read_functions = dict()

    def register_type(self, the_type, read_func):
        self.extra_read_functions[the_type] = read_func

    def set_datagram(self, datagram):
        self.datagram = datagram
        self.iter = PyDatagramIterator(self.datagram)

    def get_datagram(self):
        return self.datagram

    def set_datagram_from_blob(self, blob):
        self.set_datagram(p3d.Datagram(blob))

    def read(self, value_type, max_list_length=1000):
        builtin_converter_func = self.builtin_read_functions.get(value_type)
        if builtin_converter_func is not None:
            return builtin_converter_func()
        else:
            converter_func = self.extra_read_functions.get(value_type)
            if converter_func is not None:
                return converter_func(self)
            else:
                if type(value_type) is tuple:
                    origin_type = value_type[0]
                    if origin_type is tuple:
                        arg_types = value_type[1]
                        values = []
                        for arg_type in arg_types:
                            values.append(self.read(arg_type))
                        return tuple(values)
                    elif origin_type is list:
                        arg_type = value_type[1][0]
                        if arg_type is list:
                            raise Exception("DatagramReader does not support lists of lists.")
                        l = self.read_int16()
                        if l > max_list_length:
                            raise ExceedsListLimitException("Received list that exceeds the max list length allowed by the DatagramReader.")
                        values = []
                        for i in range(l):
                            values.append(self.read(arg_type))
                        return values
                    else:
                        raise Exception(f"Unsupported value type for DatagramReader: {value_type.__name__}")
                else:
                    raise Exception(f"Unsupported value type for DatagramReader: {value_type.__name__}")

    def read_string(self):
        return self.iter.getString()

    def read_string32(self):
        return self.iter.getString32()

    def read_bool(self):
        return self.iter.getBool()

    def read_int8(self):
        return self.iter.getInt8()

    def read_int16(self):
        return self.iter.getBeInt16()

    def read_int32(self):
        return self.iter.getBeInt32()

    def read_int64(self):
        return self.iter.getBeInt64()

    def read_float32(self):
        return self.iter.getBeFloat32()

    def read_float64(self):
        return self.iter.getBeFloat64()

    def read_blob(self):
        return self.iter.getBlob()

    def read_blob32(self):
        return self.iter.getBlob32()

    def read_vec2(self):
        return Vec2(self.read_float64(), self.read_float64())

    def read_vec3(self):
        return Vec3(self.read_float64(), self.read_float64(), self.read_float64())

    def read_vec4(self):
        return Vec4(self.read_float64(), self.read_float64(), self.read_float64(), self.read_float64())


# Gives a 32 bit hash value (shifted one right (31 bit)) that is the same across runs and devices.
def procedure_hash(name):
    h = hashlib.sha1(name.encode("utf-8"), usedforsecurity=False).digest()
    return int.from_bytes(h[:4], byteorder="big") >> 1


# -- Description --
# Main remote procedure call class used by the networking module.
# This class is likely the one you are looking for.
# -- max list length --
# Lists are supported for remote procedure calls, but to prevent attacks involving giant lists,
# there is an upper limit on the length, this can be configured, the default is small on purpose.
# -- kwargs --
# The remaining keyword arguments are passed to Peer, see the Peer class for more information.
# -- Notes --
# The first two arguments passed to a remote procedure call are connection, and time_received.
# You can disable the printing of messages on connect and disconnect via the
# print_connect, and print_disconnect booleans.
# See the networking samples on how to use this class.
class RPCPeer:
    def __init__(self, max_list_length=16, **kwargs):
        self.peer = Peer(**kwargs)

        self.max_list_length = max_list_length

        self.print_connect = True
        self.print_disconnect = True

        def default_on_connect(connection, time_connected):
            connection.rpc_peer = self
            if self.print_connect:
                print("New connection:", connection.address)
            on_connects = self.procedures.get(procedure_hash("on_connect"))
            if on_connects is not None:
                for on_connect in reversed(on_connects):
                    host_only = on_connect[3]
                    client_only = on_connect[4]
                    if self.peer.is_hosting():
                        if not client_only:
                            on_connect[2](connection, time_connected)
                    else:
                        if not host_only:
                            on_connect[2](connection, time_connected)

        def default_on_disconnect(connection, time_disconnected):
            if self.print_disconnect:
                print("Disconnected:", connection.address)
            on_disconnects = self.procedures.get(procedure_hash("on_disconnect"))
            if on_disconnects is not None:
                for on_disconnect in reversed(on_disconnects):
                    host_only = on_disconnect[3]
                    client_only = on_disconnect[4]
                    if self.peer.is_hosting():
                        if not client_only:
                            on_disconnect[2](connection, time_disconnected)
                    else:
                        if not host_only:
                            on_disconnect[2](connection, time_disconnected)

        def on_data(connection, data, time_received):
            self.rpc_on_data(connection, data, time_received)

        if self.peer.on_connect is None:
            self.peer.on_connect = default_on_connect
        if self.peer.on_disconnect is None:
            self.peer.on_disconnect = default_on_disconnect
        self.peer.on_data = on_data

        self.procedures = dict()
        self.procedures[procedure_hash("on_connect")] = []
        self.procedures[procedure_hash("on_disconnect")] = []

        self.writer = DatagramWriter()
        self.reader = DatagramReader()

    def is_using_tls(self):
        return self.peer.is_using_tls()

    def start(self, host_name, port, is_host=False, backlog=100, tls_host_name=None, socket_address_family=None):
        return self.peer.start(host_name, port, is_host=is_host, backlog=backlog, tls_host_name=tls_host_name, socket_address_family=socket_address_family)

    def stop(self):
        self.peer.stop()

    def update(self, max_events=100):
        self.peer.update(max_events=max_events)

    def is_running(self):
        return self.peer.is_running()

    def is_hosting(self):
        return self.peer.is_hosting()

    def disconnect_all(self):
        self.peer.disconnect_all()

    def connection_count(self):
        return self.peer.connection_count()

    def get_connections(self):
        return self.peer.get_connections()

    def register_type(self, the_type, write_func, read_func):
        self.writer.register_type(the_type, write_func)
        self.reader.register_type(the_type, read_func)

    def register_procedure(self, proc, host_only=False, client_only=False, prefix=None):
        func_spec = inspect.getfullargspec(proc)
        if not len(func_spec.args) >= 2:
            raise Exception(f"{proc.__name__} must have at least two arguments, connection and time_received.")
        func_args = func_spec.args[2:]
        if func_spec.args[0] == "self":
            func_args = func_args[1:]
        arg_types = []
        for func_arg in func_args:
            func_arg_type = func_spec.annotations.get(func_arg)
            if not func_arg_type is not None:
                raise Exception(f"Failed to register the '{proc.__name__}' procedure, it's missing a type annotation for the '{func_arg}' argument.")
            if type(func_arg_type) is types.GenericAlias:
                arg_types.append((typing.get_origin(func_arg_type), typing.get_args(func_arg_type)))
            else:
                arg_types.append(func_arg_type)
        proc_name = proc.__name__
        if prefix is not None:
            proc_name = prefix + "_" + proc_name
        procedure_name_hash = procedure_hash(proc_name)
        if proc.__name__ in ("on_connect", "on_disconnect"):
            p = self.procedures.get(procedure_name_hash)
            p.append((proc_name, arg_types, proc, host_only, client_only))
        else:
            if not procedure_name_hash not in self.procedures:
                raise Exception(f"{proc_name} was already registered before.")
            self.procedures[procedure_name_hash] = (proc_name, arg_types, proc, host_only, client_only)

    def __getattr__(self, name):
        def remote_procedure(*args):
            if not len(args) >= 1:
                raise Exception(f"Remote procedure call '{name}' must have at least one argument, the connection.")
            if not isinstance(args[0], Connection):
                raise Exception(f"First argument to the RPC '{name}' must be a 'Connection' type.")
            procedure_name_hash = procedure_hash(name)
            self.writer.clear()
            self.writer.write_int32(procedure_name_hash)
            connection = args[0]
            for arg in args[1:]:
                self.writer.write(arg)
            connection.send(self.writer.get_datagram().getMessage())

        return remote_procedure

    def rpc_on_data(self, connection, data, time_received):
        self.reader.set_datagram(p3d.Datagram(data))

        proc_name = None
        proc_func = None
        proc_arg_values = None
        proc_arg_types = None
        host_only = False
        client_only = False

        try:
            procedure_name_hash = self.reader.read_int32()
            proc = self.procedures.get(procedure_name_hash)
            if proc is None:
                raise Exception("Remote attempted to call a RPC that does not exist (maybe a name typo in the source code?).")
            proc_name = proc[0]
            proc_arg_types = proc[1]
            proc_func = proc[2]
            host_only = proc[3]
            client_only = proc[4]
            proc_arg_values = []
            arg_type = None
            try:
                for t in proc_arg_types:
                    arg_type = t
                    v = self.reader.read(t, max_list_length=self.max_list_length)
                    proc_arg_values.append(v)
            except ExceedsListLimitException:
                raise Exception(f"Argument with type '{arg_type}' exceeds max list size limit for procedure '{proc_name}'.")
            except Exception as e:
                raise Exception(f"Received invalid or missing argument or list/tuple exceeding max length allowed for procedure '{proc_name}', expected a '{arg_type}'.\n    {str(e)}")
        except Exception as e:
            print("WARNING: Received invalid remote procedure call, disconnecting...")
            print(e)
            connection.disconnect()

        if proc_func is not None and proc_arg_types is not None and proc_arg_values is not None:
            if len(proc_arg_values) != len(proc_arg_types):
                print("WARNING: Received invalid remote procedure call, disconnecting...")
                print(f"Received an invalid number of arguments for procedure '{proc_name}'.")
                connection.disconnect()
            else:
                if self.peer.is_hosting():
                    if not client_only:
                        proc_func(connection, time_received, *proc_arg_values)
                else:
                    if not host_only:
                        proc_func(connection, time_received, *proc_arg_values)


# Convenience attribute applied to functions to register them as remote procedure calls.
#
# @rpc(my_rpc_peer_object)
# def foo(connection, time_received, x: int):
#     print(x)
def rpc(peer, host_only=False, client_only=False):
    def wrapper(f):
        peer.register_procedure(f, host_only=host_only, client_only=client_only)
    return wrapper


# Prevent error message spam from Panda3D network module.
p3d.loadPrcFileData("", "notify-level-net fatal")



from panda3d.core import Vec3 as PandaVec3

from ursina.scripts.property_generator import generate_properties_for_class
from ursina.vec2 import Vec2


@generate_properties_for_class()
class Vec3(PandaVec3):
    def __round__(self, decimals=4):
        return Vec3(*(round(e,decimals) for e in self))


    def __repr__(self):
        return super().__repr__().replace('LVector3f', 'Vec3')


    def __iadd__(self, value):
        if len(value) % 3 == 0:
            for i in range(0, len(value), 3):
                self.add_x(value[i])
                self.add_y(value[i+1])
                self.add_z(value[i+2])
                return self

        if len(value) % 2 == 0:
            for i in range(0, len(value), 2):
                self.add_x(value[i])
                self.add_y(value[i+1])
            return self


    def __add__(self, value):
        if len(value) == 3:
            return Vec3(self[0]+value[0], self[1]+value[1], self[2]+value[2])

        if len(value) == 2:
            return Vec3(self[0]+value[0], self[1]+value[1], self[2])


    def __neg__(self):
        return Vec3(-self[0], -self[1], -self[2])


    def __sub__(self, value):
        if len(value) == 3:
            return Vec3(self[0]-value[0], self[1]-value[1], self[2]-value[2])

        if len(value) == 2:
            return Vec3(self[0]-value[0], self[1]-value[1], self[2])


    def x_getter(self):
        return self[0]
    def x_setter(self, value):
        self[0] = value

    def y_getter(self):
        return self[1]
    def y_setter(self, value):
        self[1] = value

    def z_getter(self):
        return self[2]
    def z_setter(self, value):
        self[2] = value

    def xy_getter(self):
        return Vec2(self[0], self[1])
    def xy_setter(self, value):
        self[0] = value[0]
        self[1] = value[1]

    def yx_getter(self):
        return Vec2(self[1], self[0])
    def yx_setter(self, value):
        self[1] = value[0]
        self[0] = value[1]

    def xz_getter(self):
        return Vec2(self[0], self[2])
    def xz_setter(self, value):
        self[0] = value[0]
        self[2] = value[1]

    def yz_getter(self):
        return Vec2(self[1], self[2])
    def yz_setter(self, value):
        self[1] = value[0]
        self[2] = value[1]

    def xzy_getter(self):
        return Vec3(self[0], self[2], self[1])
    def xzy_setter(self, value):
        self[0] = value[0]
        self[2] = value[1]
        self[1] = value[2]

    def X_getter(self):    # get x as int
        return int(self.x)
    def Y_getter(self):    # get y as int
        return int(self.y)
    def Z_getter(self):    # get z as int
        return int(self.z)
    def XY_getter(self):
        return Vec2(self.X, self.Y)
    def XZ_getter(self):
        return Vec2(self.X, self.Z)
    def XYZ_getter(self):
        return (self.X, self.Y, self.Z)
    def XZY_getter(self):
        return (self.X, self.Z, self.Y)


    def __mul__(self, value):
        if isinstance(value, int | float | complex):
            return Vec3(*(e*value for e in self))

        return Vec3(self[0]*value[0], self[1]*value[1], self[2]* (value[2] if len(value) > 2 else 1))


    __rmul__ = __mul__


    def __truediv__(self, value):
        if isinstance(value, int | float | complex):
            return Vec3(*(e/value for e in self))

        return Vec3(self[0]/value[0], self[1]/value[1], self[2]/value[2])

    def __abs__(self):
        return Vec3(*[abs(e) for e in self])


Vec3.zero = Vec3(0,0,0)
Vec3.one = Vec3(1,1,1)
Vec3.right = Vec3(1,0,0)
Vec3.left = Vec3(-1,0,0)
Vec3.up = Vec3(0,1,0)
Vec3.down = Vec3(0,-1,0)
Vec3.forward = Vec3(0,0,1)
Vec3.back = Vec3(0,0,-1)


if __name__ == '__main__':
    import math

    from ursinastuff import _test
    _test(Vec3(1,0,-1) * 2 == Vec3(2,0,-2))
    _test(Vec3(1,0,-1) * Vec3(1,2,3) == Vec3(1,0,-3))

    def _test_Vec3_plus_equal():
        b = Vec3(1,0,1)
        b += Vec3(0,1,0)
        return b == Vec3(1,1,1)
    _test(_test_Vec3_plus_equal)

    def _test_Vec3x_plus_equal_int():
        b = Vec3(0,0,0)
        b.x += 1
        b.y += 1
        b.z += 1
        return b == Vec3(1,1,1)

    _test(round(Vec3(1.14,2.86,1.25), 0) == Vec3(1,3,1))
    _test(abs(Vec3(1,-2,-3)) == Vec3(1,2,3))

    _test(math.isclose(Vec3(1.1, 2.5, 3.4).x, 1.1, rel_tol=1e-7))
    _test(math.isclose(Vec3(1.1, 2.5, 3.4).y, 2.5, rel_tol=1e-7))
    _test(math.isclose(Vec3(1.1, 2.5, 3.4).z, 3.4, rel_tol=1e-7))
    _test(Vec3(1.1, 2.5, 3.4).xy == Vec2(1.1, 2.5))
    _test(Vec3(1.1, 2.5, 3.4).xz == Vec2(1.1, 3.4))

    _test(Vec3(1.1, 2.5, 3.4).X == 1)
    _test(Vec3(1.1, 2.5, 3.4).Y == 2)
    _test(Vec3(1.1, 2.5, 3.4).Z == 3)
    _test(Vec3(1.1, 2.5, 3.4).XY == Vec2(1, 2))
    _test(Vec3(1.1, 2.5, 3.4).XZ == Vec2(1, 3))


from ursina import Entity, Text, camera, Button, color, mouse, Vec3
from math import floor


class ButtonList(Entity):
    def __init__(self, button_dict, button_height=1.1, width=.5, popup=False, color=Button.default_color, highlight_color=color.white33, selected_color=color.azure, font=Text.default_font, clear_selected_on_enable=True, **kwargs):
        self.clear_selected_on_enable = clear_selected_on_enable
        self.button_height = button_height
        self.width = width
        super().__init__(parent=camera.ui, position=(-(width/2), .45))

        self.text_entity = Text(parent=self, font=font, origin=(-.5,.5), text='empty', world_scale=20, z=-.1, x=.01, y=-(button_height*.25*Text.size), line_height=button_height)
        self.bg = Entity(parent=self, model='quad', origin=(-.5,.5), scale=width, color=color, collider='box')
        self.highlight = Entity(parent=self.bg, model='quad', color=highlight_color, scale=(1,self.button_height), origin=(-.5,.5), z=-.01, add_to_scene_entities=False)
        self.selection_marker = Entity(parent=self.bg, model='quad', color=selected_color, scale=(1,self.button_height), origin=(-.5,.5), z=-.02, enabled=False, add_to_scene_entities=False)
        self.button_dict = button_dict

        self.popup = popup
        if self.popup:
            self.close_button = Entity(parent=self, scale=Vec3(100,100,.1), model='quad', collider='box', visible_self=False, on_click=self.disable)

        for key, value in kwargs.items():
            setattr(self, key, value)


    @property
    def button_dict(self):
        return self._button_dict

    @button_dict.setter
    def button_dict(self, value):
        self._button_dict = value
        self.actions = list(self._button_dict.values())
        self.bg.scale_y = self.button_height * len(value) * Text.size
        # self.bg.model='circle'
        self.bg.model = 'quad'
        self.bg.origin = self.bg.origin

        self.text_entity.text = '\n'.join(self.button_dict.keys())
        self.text_entity.line_height = self.button_height
        if len(value) > 0:
            self.highlight.scale_y = 1 / len(value)
            self.selection_marker.scale_y = 1 / len(value)


    def input(self, key):
        # handle click here instead of in on_click so you can assign a custom on_click function
        if key == 'left mouse down' and self.bg.hovered:
            y = floor(-mouse.point.y * len(self.button_dict))
            y = min(y, len(self.button_dict)-1)

            action = self.actions[y]
            self.selection_marker.enabled = True
            self.selection_marker.y = self.highlight.y

            if callable(action):
                action()

            if self.popup:
                self.disable()

        if key == 'left mouse down' and not self.bg.hovered:
            self.selection_marker.enabled = False


    def update(self):
        self.highlight.enabled = mouse.hovered_entity == self.bg
        if mouse.hovered_entity == self.bg:
            y = floor(-mouse.point.y * len(self.button_dict))
            y = min(y, len(self.button_dict)-1)
            self.highlight.y = -y / len(self.button_dict)


    def on_disable(self):
        self.selection_marker.enabled = False

    def on_enable(self):
        if self.clear_selected_on_enable:
            self.selected = None


    @property
    def selected(self):
        return getattr(self, '_selected', None)
    @selected.setter
    def selected(self, value):
        self._selected = value
        if not hasattr(self, 'selection_marker'):
            return

        if not value:
            self.selection_marker.enabled = False
            return

        self.selection_marker.enabled = True
        y = list(self.button_dict.keys()).index(value)
        self.selection_marker.y = -y / len(self.button_dict)


if __name__ == '__main__':
    from ursina import Ursina, Func
    app = Ursina()

    default = Func(print, 'not yet implemented')

    def test(a=1, b=2):
        print('------:', a, b)

    button_dict = {}
    for i in range(6, 20):
        button_dict[f'button {i}'] = Func(print, i)

    bl = ButtonList(button_dict, font=Text.default_monospace_font, button_height=1.5, popup=0, clear_selected_on_enable=False)
    def input(key):
        if key == 'space':
            bl.button_dict = {
                'one' :     None,
                'two' :     default,
                'tree' :    Func(test, 3, 4),
                'four' :    Func(test, b=3, a=4),
            }
        if key == 'o':
            bl.enabled = True

    bl.selected = 'button 7'

    bl.button_dict = {}

    app.run()
